import threading
import time
from mindsdb.utilities import log

from mindsdb.utilities.log import initialize_log

from mindsdb.utilities.config import Config
from mindsdb.interfaces.storage import db
from .chatbot_task import ChatBotTask
from mindsdb.utilities.context import context as ctx


class Task(threading.Thread):
    def __init__(self, bot_record):
        threading.Thread.__init__(self)
        self.bot_record = bot_record
        self._to_stop = False

    def run(self):
        # create context and session

        ctx.set_default()
        ctx.company_id = self.bot_record.company_id
        if self.bot_record.user_class is not None:
            ctx.user_class = self.bot_record.user_class

        from mindsdb.api.mysql.mysql_proxy.controllers.session_controller import SessionController
        session = SessionController()

        # TODO check deleted, raise errors
        # TODO checks on delete predictor/ project/ integration
        model_name = self.bot_record.model_name
        project_name = db.Project.query.get(self.bot_record.project_id).name

        database_name = db.Integration.query.get(self.bot_record.database_id).name

        task = ChatBotTask(session,
                           database=database_name,
                           project_name=project_name,
                           model_name=model_name,
                           params=self.bot_record.params)

        while True:
            try:
                task.run()
            except Exception as e:
                log.logger.error(e)

            if self._to_stop:
                return
            log.logger.debug('running ' + self.name)
            time.sleep(7)

    def stop(self):
        self._to_stop = True


class ChatBotMonitor:
    def __init__(self):
        self._active_bots = {}

    def start(self):
        config = Config()
        db.init()
        initialize_log(config, 'jobs', wrap_print=True)
        self.config = config

        while True:
            try:
                self.check_bots()

            except (SystemExit, KeyboardInterrupt):
                self.stop_all_bots()
                raise
            except Exception as e:
                log.logger.error(e)

            db.session.rollback()  # disable cache
            time.sleep(1)

    def stop_all_bots(self):
        active_bots = list(self._active_bots.keys())
        for bot_id in active_bots:
            self.stop_bot(bot_id)

    def check_bots(self):
        allowed_bots = []

        # check new bots to start
        for bot in db.ChatBots.query.all():
            bot_id = bot.id
            allowed_bots.append(bot_id)

            if bot_id not in self._active_bots:
                # start new bot
                thread = Task(bot)
                thread.start()

                self._active_bots[bot.id] = thread

        # check old bots to stop
        active_bots = list(self._active_bots.keys())
        for bot_id in active_bots:
            if bot_id not in allowed_bots:
                self.stop_bot(bot_id)

    def stop_bot(self, bot_id):
        self._active_bots[bot_id].stop()
        del self._active_bots[bot_id]


def start(verbose=False):
    is_cloud = Config().get('cloud', False)
    if is_cloud is True:
        # Chatbots are disabled on cloud
        return

    monitor = ChatBotMonitor()
    monitor.start()


if __name__ == '__main__':
    start()



import pandas as pd

from typing import List

from mindsdb.integrations.libs.api_handler import APITable
from mindsdb.integrations.utilities.sql_utils import extract_comparison_conditions
from mindsdb.utilities.log import get_log

from mindsdb_sql.parser import ast

logger = get_log("integrations.InfluxDB_handler")

class InfluxDBTables(APITable):
    """InfluxDB Tables implementation"""

    def select(self, query: ast.Select) -> pd.DataFrame:
        """Pulls data from the InfluxDB "query" API endpoint
        Parameters
        ----------
        query : ast.Select
           Given SQL SELECT query
        Returns
        -------
        pd.DataFrame of particular InfluxDB table matching the query
        Raises
        ------
        ValueError
            If the query contains an unsupported condition
        """
        conditions = extract_comparison_conditions(query.where)

       
        order_by_conditions = {}

        if query.order_by and len(query.order_by) > 0:
            order_by_conditions["columns"] = []
            order_by_conditions["ascending"] = []

            for an_order in query.order_by:
                if an_order.field.parts[0] != "":
                    next    
                if an_order.field.parts[1] in self.get_columns():
                    order_by_conditions["columns"].append(an_order.field.parts[1])

                    if an_order.direction == "ASC":
                        order_by_conditions["ascending"].append(True)
                    else:
                        order_by_conditions["ascending"].append(False)
                else:
                    raise ValueError(
                        f"Order by unknown column {an_order.field.parts[1]}"
                    )

        influxdb_tables_df = self.handler.call_influxdb_tables()

        selected_columns = []
        for target in query.targets:
            if isinstance(target, ast.Star):
                selected_columns = self.get_columns()
                break
            elif isinstance(target, ast.Identifier):
                selected_columns.append(target.parts[-1])
            else:
                raise ValueError(f"Unknown query target {type(target)}")


        if len(influxdb_tables_df) == 0:
            influxdb_tables_df = pd.DataFrame([], columns=selected_columns)
        else:
            influxdb_tables_df.columns = self.get_columns()
            for col in set(influxdb_tables_df.columns).difference(set(selected_columns)):
                influxdb_tables_df = influxdb_tables_df.drop(col, axis=1)

            if len(order_by_conditions.get("columns", [])) > 0:
                influxdb_tables_df = influxdb_tables_df.sort_values(
                    by=order_by_conditions["columns"],
                    ascending=order_by_conditions["ascending"],
                )

        if query.limit:
            influxdb_tables_df = influxdb_tables_df.head(query.limit.value)

        return influxdb_tables_df

    def get_columns(self) -> List[str]:
        """Gets all columns to be returned in pandas DataFrame responses
        Returns
        -------
        List[str]
            List of columns
        """
        influxdb_df_fields = self.handler.call_influxdb_tables()

        return list(influxdb_df_fields.columns)



import struct
from typing import List, Any, BinaryIO, Sequence, Dict

from mindsdb.api.postgres.postgres_proxy.utilities import strip_null_byte
from mindsdb.utilities.log import get_log
import time
from mindsdb.api.postgres.postgres_proxy.postgres_packets.postgres_fields import PostgresField


from mindsdb.api.postgres.postgres_proxy.postgres_packets.postgres_message_identifiers import \
    PostgresBackendMessageIdentifier, PostgresFrontendMessageIdentifier, PostgresAuthType


class PostgresEmptyDataException(Exception):
    pass


class UnsupportedSSLRequest(Exception):
    pass


class UnsupportedPostgresAuthException(Exception):
    pass


class UnsupportedPostgresMessageType(Exception):
    pass


class PostgresPacketReader:
    def __init__(self, buffer: BinaryIO):
        from mindsdb.api.postgres.postgres_proxy.postgres_packets.postgres_message_formats import FE_MESSAGE_MAP, \
            SUPPORTED_AUTH_TYPES
        self.fe_message_map = FE_MESSAGE_MAP
        self.supported_auth_types = SUPPORTED_AUTH_TYPES
        self.buffer = buffer
        self.logger = get_log('postgres_proxy')

    def read_byte(self):
        return self.read_bytes(1)

    def read_bytes(self, n):
        data = self.buffer.read(n)
        self.logger.debug("received data:")
        self.logger.debug(data.hex())
        if not data:
            raise PostgresEmptyDataException("Expected data inside of buffer when performing read_bytes")
        else:
            return data

    def read_string(self):
        result = b""
        while True:
            b = self.read_byte()
            if b == b"\x00":
                break
            result = result + b
        return result

    def read_bytes_timeout(self, n, timeout=60):
        cur = time.time()
        end = cur + timeout
        # This is a reason to switch from socketserver to asyncio
        while cur < end:
            cur = time.time()
            data = self.buffer.read(n)
            if not data:
                pass
            else:
                return data
        raise PostgresEmptyDataException("Expected data inside of buffer when performing read_bytes")

    def read_int16(self):
        data = self.read_bytes(2)
        return struct.unpack("!h", data)[0]

    def read_int32(self):
        data = self.read_bytes(4)
        return struct.unpack("!i", data)[0]

    def read_parameters(self, n):
        data = self.read_bytes(n)
        return data.split(b'\x00')

    def read_verify_ssl_request(self):
        self.logger.debug("reading ssl")
        length = self.read_int32()
        code = self.read_int32()
        if length != 8 and code != 80877103:
            raise UnsupportedSSLRequest("Code %s of len %s" % (code, length))

    def read_startup_message(self) -> Dict[bytes, bytes]:
        self.logger.debug("reading startup message")
        length = self.read_int32()
        version = self.read_int32()
        major_version = version >> 16
        minor_version = version & 0xffff
        message = self.read_parameters(length - 8)
        self.logger.debug('PSQL Startup Message %d.%d : %s' % (major_version, minor_version, message))
        parameters = {}
        while len(message) != 0:
            key = message.pop(0)
            value = message.pop(0)
            parameters[key] = value
        parameters[b"major_version"] = major_version
        parameters[b"minor_version"] = minor_version
        return parameters

    def read_authentication(self, encoding=None):
        try:
            auth_type = self.read_byte()
        except PostgresEmptyDataException:
            # No authentication parameters specified. Which is fine if we're local on a mindsdbuser
            return ''
        try:
            auth_type = PostgresAuthType(auth_type)
        except Exception as e:
            raise UnsupportedPostgresAuthException(e)
        if auth_type not in self.supported_auth_types:
            raise UnsupportedPostgresAuthException("%s is not a supported auth type identifier" % auth_type)
        length = self.read_int32()
        password = strip_null_byte(self.read_bytes(length - 4), encoding=encoding)  # password. Do something with later. We read to clear buffer.
        return password

    def read_message(self):
        try:
            message_type = self.read_byte()
        except PostgresEmptyDataException:
            self.logger.warn("Postgres Proxy: Received empty data string")
            return None
        try:
            message_type = PostgresFrontendMessageIdentifier(message_type)
        except Exception as e:
            raise UnsupportedPostgresMessageType(
                "%s is not a supported frontend message identifier:\n%s" % (message_type, str(e)))

        if message_type in self.fe_message_map:
            self.logger.debug("reading message type %s" % str(message_type.name))
            return self.fe_message_map[message_type]().read(self)
        else:
            raise UnsupportedPostgresMessageType(
                "%s is not a supported frontend message identifier" % message_type.value)


class PostgresPacketBuilder:
    identifier: bytes
    pack_string: str
    length: int
    pack_args: List[Any]

    def __init__(self):
        self.logger = get_log('postgres_proxy')
        self.reset()

    def reset(self):
        self.identifier = b''
        self.pack_string = ''
        self.pack_args = []
        self.length = 0

    def set_identifier(self, message_identifier: PostgresBackendMessageIdentifier):
        self.identifier = message_identifier.value
        return self

    def add_length(self, length: int):
        self.length += length
        return self

    def write_char(self, c, write_file: BinaryIO):
        pack = "!c"
        packed_binary = struct.pack(pack, c)
        write_file.write(packed_binary)

    def write(self, write_file: BinaryIO):
        if len(self.identifier) == 0:
            raise Exception("Can't write without identifier.")
        pack = "!c"

        # Send length
        # Ensures 'i' in pack and self.length in 2nd position of respective variables.
        # Also adds 4 to length to include int32 rep of length
        self.length += 4
        pack += "i"
        self.pack_args = [self.length] + self.pack_args

        pack += self.pack_string
        self.logger.debug("writing:")
        self.logger.debug("pack string: %s" % self.pack_string)
        self.logger.debug("identifier: %s" % self.pack_string)
        self.logger.debug("pack args:")
        for arg in self.pack_args:
            self.logger.debug("arg: %s" % str(arg))
        packed_binary = struct.pack(pack, self.identifier, *self.pack_args)
        write_file.write(packed_binary)

    def add_char(self, s: bytes):
        self.pack_string += 'c'
        if len(s) != 1:
            raise Exception("Char must be of length 1 in add_char")
        self.pack_args.append(s)
        return self.add_length(1)

    def add_string(self, s: bytes):
        s = s + b'\x00'
        self.pack_string += str(len(s)) + 's'
        self.pack_args.append(s)
        return self.add_length(len(s))

    def add_int32(self, i):
        self.pack_string += 'i'
        self.pack_args.append(i)
        return self.add_length(4)

    def add_int16(self, h):
        self.pack_string += 'h'
        self.pack_args.append(h)
        return self.add_length(2)

    def add_bytes(self, b: bytes):
        if len(b) == 1:
            self.pack_string += 's'
        else:
            self.pack_string += str(len(b)) + 's'
        self.pack_args.append(b)
        return self.add_length(len(b))

    def add_field(self, field: PostgresField):
        return self.add_string(field.name.encode()) \
            .add_int32(field.table_id) \
            .add_int16(field.column_id) \
            .add_int32(field.object_id) \
            .add_int16(field.dt_size) \
            .add_int32(field.type_modifier) \
            .add_int16(field.format_code)

    def add_fields(self, fields: Sequence[PostgresField]):
        for field in fields:
            self.add_field(field)
        return self

    def add_column_value(self, val: bytes):
        if val == b'NULL':
            self.add_int32(-1)
            return self

        self.add_int32(len(val))
        self.add_bytes(val)
        return self

    def add_row(self, row: Sequence[bytes]):
        for val in row:
            self.add_column_value(val)
        return self



import ray


class RayConnection:
    def __init__(self, addr=None, **kwargs):
        ray.init(address=addr, **kwargs)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback):
        ray.shutdown()



from mindsdb.utilities.config import Config
from mindsdb.interfaces.storage import db
from mindsdb.interfaces.database.projects import ProjectController

from mindsdb.utilities.context import context as ctx


class ChatBotController:

    def add_chatbot(self, name, project_name, model_name, database_id, params):
        is_cloud = Config().get('cloud', False)
        if is_cloud is True:
            raise Exception('Chatbots are disabled on cloud')

        if project_name is None:
            project_name = 'mindsdb'
        project_controller = ProjectController()
        project = project_controller.get(name=project_name)

        bot = db.ChatBots.query.filter(
            db.ChatBots.company_id == ctx.company_id,
            db.ChatBots.name == name,
            db.ChatBots.project_id == project.id
        ).first()

        if bot is not None:
            raise Exception(f'Chat bot already exists: {name}')

        bot = db.ChatBots(
            company_id=ctx.company_id,
            name=name,
            project_id=project.id,
            model_name=model_name,
            database_id=database_id,
            params=params,
        )
        db.session.add(bot)
        db.session.commit()

    def delete_chatbot(self, name, project_name):
        project_controller = ProjectController()
        project = project_controller.get(name=project_name)

        bot = db.ChatBots.query.filter(
            db.ChatBots.company_id == ctx.company_id,
            db.ChatBots.name == name,
            db.ChatBots.project_id == project.id
        ).first()

        if bot is None:
            raise Exception(f'Chat bot not found: {name}')

        db.session.delete(bot)
        db.session.commit()



import time

import pandas as pd
from langchain.text_splitter import RecursiveCharacterTextSplitter
from langchain.vectorstores import Chroma

from mindsdb.utilities.log import get_log

from .settings import (
    DEFAULT_EMBEDDINGS_MODEL,
    df_to_documents,
    get_chroma_settings,
    load_embeddings_model,
)

logger = get_log(__name__)


class Ingestor:
    def __init__(self, args: dict, df: pd.DataFrame):
        self.args = args
        self.df = df
        self.embeddings_model_name = args.get(
            "embeddings_model_name", DEFAULT_EMBEDDINGS_MODEL
        )
        self.persist_directory = args["chromadb_storage_path"]

        self.chroma_settings = get_chroma_settings(self.persist_directory)

    def split_documents(self):
        # Load documents and split in chunks
        logger.info(f"Loading documents from input data")

        text_splitter = RecursiveCharacterTextSplitter(chunk_size=500, chunk_overlap=50)
        documents = df_to_documents(
            df=self.df, page_content_columns=self.args["context_columns"]
        )
        texts = text_splitter.split_documents(documents)
        logger.info(f"Loaded {len(documents)} documents from input data")
        logger.info(f"Split into {len(texts)} chunks of text (max. 500 tokens each)")

        return texts

    def embeddings_to_vectordb(self):
        start_time = time.time()

        # Load documents and split in chunks
        texts = self.split_documents()

        # Load embeddings model
        embeddings_model = load_embeddings_model(self.embeddings_model_name)

        logger.info(f"Creating vectorstore from documents")

        # Create and store locally vectorstore
        db = Chroma.from_documents(
            texts,
            embedding=embeddings_model,
            persist_directory=self.persist_directory,
            client_settings=self.chroma_settings,
        )
        db.persist()
        db = None
        end_time = time.time()
        elapsed_time = end_time - start_time

        logger.info(
            "Fished creating vectorstore from documents. It took: {elapsed_time/60} minutes"
        )

        logger.info("Finished creating vectorstore from documents.")
        logger.info(f"Elapsed time: {round(elapsed_time / 60)} minutes")



__title__ = 'MindsDB'
__package_name__ = 'mindsdb'
__version__ = '23.6.4.0'
__description__ = "MindsDB server, provides server capabilities to mindsdb native python library"
__email__ = "jorge@mindsdb.com"
__author__ = 'MindsDB Inc'
__github__ = 'https://github.com/mindsdb/mindsdb'
__pypi__ = 'https://pypi.org/project/mindsdb/'
__license__ = 'GPL-3.0'
__copyright__ = 'Copyright 2018- mindsdb'



import pandas as pd
from typing import Dict

from binance.spot import Spot

from mindsdb.integrations.handlers.binance_handler.binance_tables import BinanceAggregatedTradesTable
from mindsdb.integrations.libs.api_handler import APIHandler
from mindsdb.integrations.libs.response import (
    HandlerStatusResponse as StatusResponse,
    HandlerResponse as Response,
)
from mindsdb.utilities import log
from mindsdb_sql import parse_sql

_BASE_BINANCE_US_URL = 'https://api.binance.us'


class BinanceHandler(APIHandler):
    """A class for handling connections and interactions with the Binance API.

    Attributes:
        client (binance.spot.Spot): The `binance.spot.Spot` object for interacting with the Binance API.
        api_key (str): API key, only required for user data endpoints.
        api_secret (str): API secret, only required for user data endpoints.
        is_connected (bool): Whether or not the API client is connected to Binance.

    """

    def __init__(self, name: str = None, **kwargs):
        """Registers all API tables and prepares the handler for an API connection.

        Args:
            name: (str): The handler name to use
        """
        super().__init__(name)
        self.api_key = None
        self.api_secret = None

        args = kwargs.get('connection_data', {})
        if 'api_key' in args:
            self.api_key = args['api_key']
        if 'api_secret' in args:
            self.api_secret = args['api_secret']

        self.client = None
        self.is_connected = False

        aggregated_trade_data = BinanceAggregatedTradesTable(self)
        self._register_table('aggregated_trade_data', aggregated_trade_data)

    def connect(self) -> Spot:
        """Creates a new Binance Spot API client if needed and sets it as the client to use for requests.

        Returns newly created Binance Spot API client, or current client if already set.
        """
        if self.is_connected is True and self.client:
            return self.client

        if self.api_key and self.api_secret:
            self.client = Spot(key=self.api_key, secret=self.api_secret, base_url=_BASE_BINANCE_US_URL)
        else:
            self.client = Spot(base_url=_BASE_BINANCE_US_URL)

        self.is_connected = True
        return self.client

    def check_connection(self) -> StatusResponse:
        """Checks connection to Binance API by sending a ping request.

        Returns StatusResponse indicating whether or not the handler is connected.
        """

        response = StatusResponse(False)

        try:
            client = self.connect()
            client.ping()
            response.success = True

        except Exception as e:
            log.logger.error(f'Error connecting to Binance API: {e}!')
            response.error_message = e

        self.is_connected = response.success
        return response

    def _get_klines(self, params: Dict = None) -> pd.DataFrame:
        """Gets aggregate trade data for a symbol based on given parameters

        Returns results as a pandas DataFrame.

        Args:
            params (Dict): Trade data params (symbol, interval, limit, start_time, end_time)
        """
        if 'symbol' not in params:
            raise ValueError('Missing "symbol" param to fetch trade data for.')
        if 'interval' not in params:
            raise ValueError('Missing "interval" param (1m, 1d, etc).')

        client = self.connect()
        symbol = params['symbol']
        interval = params['interval']
        limit = params['limit'] if 'limit' in params else BinanceAggregatedTradesTable.DEFAULT_AGGREGATE_TRADE_LIMIT
        start_time = params['start_time'] if 'start_time' in params else None
        end_time = params['end_time'] if 'end_time' in params else None
        raw_klines = client.klines(
            symbol,
            interval,
            limit=limit,
            startTime=start_time,
            endTime=end_time)

        open_time_i = 0
        close_time_i = 6
        for i in range(len(raw_klines)):
            # To train we need timestamps to be in seconds since Unix epoch and Binance returns it in ms.
            raw_klines[i][open_time_i] /= 1000
            raw_klines[i][close_time_i] /= 1000

        df = pd.DataFrame(raw_klines)
        df.insert(0, 'symbol', [symbol] * len(raw_klines), True)
        # Remove last unnecessary column (unused API field)
        if len(raw_klines) > 0:
            num_cols = len(df.columns)
            df = df.drop(df.columns[[num_cols - 1]], axis=1)
        return df

    def native_query(self, query: str = None) -> Response:
        ast = parse_sql(query, dialect='mindsdb')
        return self.query(ast)

    def call_binance_api(self, method_name: str = None, params: Dict = None) -> pd.DataFrame:
        """Calls the Binance API method with the given params.

        Returns results as a pandas DataFrame.

        Args:
            method_name (str): Method name to call (e.g. klines)
            params (Dict): Params to pass to the API call
        """
        if method_name == 'klines':
            return self._get_klines(params)
        raise NotImplementedError('Method name {} not supported by Binance API Handler'.format(method_name))



from textwrap import dedent
from collections import OrderedDict

from pandas import DataFrame
from sqlalchemy import String

from sqlalchemy.sql import text, bindparam

import teradatasql
import teradatasqlalchemy.dialect as teradata_dialect

from mindsdb_sql import parse_sql
from mindsdb_sql.parser.ast.base import ASTNode
from mindsdb_sql.render.sqlalchemy_render import SqlalchemyRender

from mindsdb.utilities import log

from mindsdb.integrations.libs.base import DatabaseHandler
from mindsdb.integrations.libs.response import (
    HandlerStatusResponse as StatusResponse,
    HandlerResponse as Response,
    RESPONSE_TYPE
)
from mindsdb.integrations.libs.const import HANDLER_CONNECTION_ARG_TYPE as ARG_TYPE


class TeradataHandler(DatabaseHandler):
    """
    This handler handles connection and execution of the Teradata statements.
    """

    name = 'teradata'

    def __init__(self, name: str, connection_data: dict, **kwargs):
        super().__init__(name)

        self.dialect = 'teradata'
        self.parser = parse_sql
        self.connection_data = connection_data
        self.renderer = SqlalchemyRender(teradata_dialect.TeradataDialect)

        self.host = self.connection_data.get('host')
        self.database = self.connection_data.get('database')

        self.connection = None
        self.is_connected = False

    def __del__(self):
        if self.is_connected is True:
            self.disconnect()

    def connect(self):
        """
        Handles the connection to a Teradata database insance.
        """

        if self.is_connected is True:
            return self.connection

        connection = teradatasql.connect(
            **self.connection_data
        )

        self.is_connected = True
        self.connection = connection
        return self.connection

    def disconnect(self):
        """
        Disconnects from the Teradata database
        """

        if self.is_connected is True:
            self.connection.close()
            self.is_connected = False

    def check_connection(self) -> StatusResponse:
        """
        Check the connection of the Teradata database
        :return: success status and error message if error occurs
        """

        response = StatusResponse(False)
        need_to_close = self.is_connected is False

        try:
            connection = self.connect()
            with connection.cursor() as cur:
                cur.execute('SELECT 1 FROM (SELECT 1 AS "dual") AS "dual"')
            response.success = True
        except teradatasql.Error as e:
            log.logger.error(f'Error connecting to Teradata {self.host}, {e}!')
            response.error_message = e

        if response.success is True and need_to_close:
            self.disconnect()
        if response.success is False and self.is_connected is True:
            self.is_connected = False

        return response

    def native_query(self, query: str) -> Response:
        """
        Receive SQL query and runs it
        :param query: The SQL query to run in Teradata
        :return: returns the records from the current recordset
        """

        need_to_close = self.is_connected is False

        connection = self.connect()
        with connection.cursor() as cur:
            try:
                cur.execute(query)
                if not cur.description:
                    response = Response(RESPONSE_TYPE.OK)
                else:
                    result = cur.fetchall()
                    response = Response(
                        RESPONSE_TYPE.TABLE,
                        DataFrame(
                            result,
                            columns=[x[0] for x in cur.description]
                        )
                    )
                connection.commit()
            except Exception as e:
                log.logger.error(f'Error running query: {query} on {self.host}!')
                response = Response(
                    RESPONSE_TYPE.ERROR,
                    error_code=0,
                    error_message=str(e)
                )
                connection.rollback()

        if need_to_close is True:
            self.disconnect()

        return response

    def query(self, query: ASTNode) -> Response:
        """
        Retrieve the data from the SQL statement with eliminated rows that dont satisfy the WHERE condition
        """

        query_str = self.renderer.get_string(query, with_failback=True)
        return self.native_query(query_str)

    def get_tables(self) -> Response:
        """
        List all tables in Teradata in the current database
        """

        return self.native_query(
            str(text(f"""
            SELECT DataBaseName,
                   TableName,
                   TableKind
            FROM DBC.TablesV
            WHERE DatabaseName = :database
            AND (TableKind = 'T'
                OR TableKind = 'O'
                OR TableKind = 'Q')
            """).bindparams(
                bindparam('database', value=self.database, type_=String)
            ).compile(compile_kwargs={"literal_binds": True}))
        )

    def get_columns(self, table_name: str) -> Response:
        """
        List all columns in a table in Teradata in the current schema
        :param table_name: the table name for which to list the columns
        :return: returns the columns in the table
        """

        return self.native_query(
            str(text(f"""
            SELECT ColumnName AS "Field",
                   ColumnType AS "Type"
            FROM DBC.ColumnsV
            WHERE DatabaseName (NOT CASESPECIFIC) = :database
            AND TableName (NOT CASESPECIFIC) = :table_name
            """).bindparams(
                bindparam('database', value=self.database, type_=String),
                bindparam('table_name', value=table_name, type_=String)
            ).compile(compile_kwargs={"literal_binds": True}))
        )


connection_args = OrderedDict(
    host={
        'type': ARG_TYPE.STR,
        'description': 'The IP address/host name of the Teradata instance host.'
    },
    user={
        'type': ARG_TYPE.STR,
        'description': 'Specifies the user name.'
    },
    password={
        'type': ARG_TYPE.STR,
        'description': 'Specifies the password for the user.'
    },
    database={
        'type': ARG_TYPE.STR,
        'description': 'Specifies the initial database to use after logon, instead of the default database.'
    },
    dbs_port={
        'type': ARG_TYPE.STR,
        'description': 'The port number of the Teradata instance.'
    },
    encryptdata={
        'type': ARG_TYPE.STR,
        'description': 'Controls encryption of data exchanged between the driver and the database.'
    },
    https_port={
        'type': ARG_TYPE.STR,
        'description': 'Specifies the database port number for HTTPS/TLS connections.'
    },
    sslca={
        'type': ARG_TYPE.STR,
        'description': 'Specifies the file name of a PEM file that contains Certificate Authority (CA) certificates.'
    },
    sslcapath={
        'type': ARG_TYPE.STR,
        'description': 'Specifies a directory of PEM files that contain Certificate Authority (CA) certificates.'
    },
    sslcipher={
        'type': ARG_TYPE.STR,
        'description': 'Specifies the TLS cipher for HTTPS/TLS connections.'
    },
    sslmode={
        'type': ARG_TYPE.STR,
        'description': 'Specifies the SSL mode for HTTPS/TLS connections.'
    },
    sslprotocol={
        'type': ARG_TYPE.STR,
        'description': 'Specifies the TLS protocol for HTTPS/TLS connections.'
    },
    tmode={
        'type': ARG_TYPE.STR,
        'description': 'Specifies the Teradata transaction mode.'
    },
    logmech={
        'type': ARG_TYPE.STR,
        'description': 'Specifies the login authentication method.'
    },
    logdata={
        'type': ARG_TYPE.STR,
        'description': 'Specifies extra data for the chosen logon authentication method.'
    },
    browser={
        'type': ARG_TYPE.STR,
        'description': 'Specifies the command to open the browser for Browser Authentication.'
    }
)

connection_args_example = OrderedDict(
    host='192.168.0.41',
    user='dbc',
    password='dbc',
    database='HR'
)



"""chatbot

Revision ID: aaecd7012a78
Revises: b8be148dbc85
Create Date: 2023-05-18 10:53:35.256508

"""
from alembic import op
import sqlalchemy as sa
import mindsdb.interfaces.storage.db  # noqa


# revision identifiers, used by Alembic.
revision = 'aaecd7012a78'
down_revision = '6d748f2c7b0b'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        'chat_bots',
        sa.Column('id', sa.Integer(), nullable=False),
        sa.Column('company_id', sa.Integer(), nullable=True),
        sa.Column('user_class', sa.Integer(), nullable=True),
        sa.Column('name', sa.String(), nullable=False),
        sa.Column('project_id', sa.Integer(), nullable=False),
        sa.Column('model_name', sa.String(), nullable=False),
        sa.Column('database_id', sa.Integer(), nullable=False),
        sa.Column('params', sa.JSON(), nullable=True),
        sa.Column('updated_at', sa.DateTime(), nullable=True),
        sa.Column('created_at', sa.DateTime(), nullable=True),
        sa.PrimaryKeyConstraint('id')
    )


def downgrade():

    op.drop_table('chat_bots')
    # ### end Alembic commands ###



from mindsdb.integrations.libs.const import HANDLER_TYPE

try:
    from .lightwood_handler.lightwood_handler import LightwoodHandler as Handler
    import_error = None
except Exception as e:
    Handler = None
    import_error = e
from .lightwood_handler.__about__ import __version__ as version


title = 'Lightwood'
name = 'lightwood'
type = HANDLER_TYPE.ML
permanent = True


__all__ = [
    'Handler', 'version', 'name', 'type', 'title', 'import_error'
]



import os
import json
from copy import deepcopy
from pathlib import Path

from mindsdb.utilities.fs import create_directory, get_or_create_data_dir


def _merge_key_recursive(target_dict, source_dict, key):
    if key not in target_dict:
        target_dict[key] = source_dict[key]
    elif not isinstance(target_dict[key], dict) or not isinstance(source_dict[key], dict):
        target_dict[key] = source_dict[key]
    else:
        for k in list(source_dict[key].keys()):
            _merge_key_recursive(target_dict[key], source_dict[key], k)


def _merge_configs(original_config, override_config):
    original_config = deepcopy(original_config)
    for key in list(override_config.keys()):
        _merge_key_recursive(original_config, override_config, key)
    return original_config


config = None
config_mtime = -1


class Config():
    def __init__(self):
        # initialize once
        global config, config_mtime
        self.config_path = os.environ.get('MINDSDB_CONFIG_PATH', 'absent')
        self.use_docker_env = os.environ.get('MINDSDB_DOCKER_ENV', False)
        if self.use_docker_env:
            self.use_docker_env = True

        if Path(self.config_path).is_file():
            current_config_mtime = os.path.getmtime(self.config_path)
            if config_mtime != current_config_mtime:
                config = self.init_config()
                config_mtime = current_config_mtime

        if config is None:
            config = self.init_config()

        self._config = config

    def init_config(self):
        if self.config_path == 'absent':
            self._override_config = {}
        else:
            with open(self.config_path, 'r') as fp:
                self._override_config = json.load(fp)

        # region define storage dir
        if 'storage_dir' in self._override_config:
            root_storage_dir = self._override_config['storage_dir']
            os.environ['MINDSDB_STORAGE_DIR'] = root_storage_dir
        elif os.environ.get('MINDSDB_STORAGE_DIR') is not None:
            root_storage_dir = os.environ['MINDSDB_STORAGE_DIR']
        else:
            root_storage_dir = get_or_create_data_dir()
            os.environ['MINDSDB_STORAGE_DIR'] = root_storage_dir
        # endregion

        if os.path.isdir(root_storage_dir) is False:
            os.makedirs(root_storage_dir)

        if 'storage_db' in self._override_config:
            os.environ['MINDSDB_DB_CON'] = self._override_config['storage_db']
        elif os.environ.get('MINDSDB_DB_CON', '') == '':
            os.environ['MINDSDB_DB_CON'] = 'sqlite:///' + os.path.join(root_storage_dir,
                                                                       'mindsdb.sqlite3.db') + '?check_same_thread=False&timeout=30'

        paths = {
            'root': os.environ['MINDSDB_STORAGE_DIR']
        }

        # content - temporary storage for entities
        paths['content'] = os.path.join(paths['root'], 'content')
        # storage - persist storage for entities
        paths['storage'] = os.path.join(paths['root'], 'storage')
        paths['static'] = os.path.join(paths['root'], 'static')
        paths['tmp'] = os.path.join(paths['root'], 'tmp')
        paths['log'] = os.path.join(paths['root'], 'log')
        paths['cache'] = os.path.join(paths['root'], 'cache')

        for path_name in paths:
            create_directory(paths[path_name])

        api_host = "127.0.0.1" if not self.use_docker_env else "0.0.0.0"
        self._default_config = {
            'permanent_storage': {
                'location': 'local'
            },
            'storage_dir': os.environ['MINDSDB_STORAGE_DIR'],
            'paths': paths,
            'auth': {
                'http_auth_enabled': False,
                'username': 'mindsdb',
                'password': ''
            },
            "log": {
                "level": {
                    "console": "INFO",
                    "file": "DEBUG",
                    "db": "WARNING"
                }
            },
            "gui": {
                "autoupdate": True
            },
            "debug": False,
            "environment": "local",
            "integrations": {},
            "api": {
                "http": {
                    "host": api_host,
                    "port": "47334"
                },
                "mysql": {
                    "host": api_host,
                    "password": "",
                    "port": "47335",
                    "database": "mindsdb",
                    "ssl": True
                },
                "mongodb": {
                    "host": api_host,
                    "port": "47336",
                    "database": "mindsdb"
                },
                "postgres": {
                    "host": api_host,
                    "port": "55432",
                    "database": "mindsdb"
                }
            },
            "cache": {
                "type": "local"
            }
        }

        return _merge_configs(self._default_config, self._override_config)

    def __getitem__(self, key):
        return self._config[key]

    def get(self, key, default=None):
        return self._config.get(key, default)

    def get_all(self):
        return self._config

    def update(self, data: dict):
        config_path = Path(self.config_path)
        if config_path.is_file() is False:
            config_path.write_text('{}')

        with open(self.config_path, 'r') as fp:
            config_data = json.load(fp)

        config_data = _merge_configs(config_data, data)

        with open(self.config_path, 'wt') as fp:
            fp.write(json.dumps(config_data, indent=4))

        self.init_config()

    @property
    def paths(self):
        return self._config['paths']



import os.path
import json
import pandas as pd
import pytz
from tzlocal import get_localzone
from datetime import datetime

from google.auth.transport.requests import Request
from google.oauth2.credentials import Credentials
from google_auth_oauthlib.flow import InstalledAppFlow
from googleapiclient.discovery import Resource
from googleapiclient.discovery import build
from googleapiclient.errors import HttpError
from mindsdb_sql import parse_sql

from mindsdb.utilities import log
from mindsdb.integrations.handlers.google_fit_handler.google_fit_tables import GoogleFitTable
from mindsdb.integrations.libs.api_handler import APIHandler,FuncParser
from mindsdb.integrations.libs.response import (
    HandlerStatusResponse as StatusResponse,
    HandlerResponse as Response,
)

SCOPES = ['https://www.googleapis.com/auth/fitness.activity.read']
DATE_FORMAT = '%Y-%m-%d'
class GoogleFitHandler(APIHandler):

    def __init__(self, name: str = None, **kwargs):
        super().__init__(name)
        args = kwargs.get('connection_data', {})
        self.connection_args = {}
        self.credentials_path = None
        if 'service_account_file' in args:
            if os.path.isfile(args['service_account_file']) is False:
                raise Exception("service_account_file must be a path to the credentials.json file")
            self.credentials_path = args['service_account_file']
        elif 'service_account_json' in args:
            self.connection_args = args['service_account_json']
            if not isinstance(self.connection_args, dict) or (('redirect_uris' not in self.connection_args.keys()) and len(self.connection_args) != 6) or ('redirect_uris' in self.connection_args.keys()) and len(self.connection_args) != 7:
                raise Exception("service_account_json has to be a dictionary with all 6 required fields")
            self.connection_args['redirect_uris'] = ['http://localhost']
            self.credentials_path = 'mindsdb/integrations/handlers/google_fit_handler/credentials.json'
        else:
            raise Exception('Connection args have to content ether service_account_file or service_account_json')
        

        self.api = None
        self.is_connected = False

        aggregated_data = GoogleFitTable(self)
        self._register_table('aggregated_data', aggregated_data)

    def connect(self) -> Resource:
        if self.is_connected is True and self.api:
            return self.api
        if self.connection_args:
            credentialDict = {"installed":self.connection_args}
            f = open(self.credentials_path, "w")
            f.write(json.dumps(credentialDict).replace(" ", ""))
            f.close()
            
        creds = None

        if os.path.isfile('mindsdb/integrations/handlers/google_fit_handler/token.json'):
            creds = Credentials.from_authorized_user_file('mindsdb/integrations/handlers/google_fit_handler/token.json', SCOPES)
        if not creds or not creds.valid:
            if creds and creds.expired and creds.refresh_token:
                creds.refresh(Request())
            else:
                flow = InstalledAppFlow.from_client_secrets_file(
                    self.credentials_path, SCOPES)
                creds = flow.run_local_server(port=0)
            with open('mindsdb/integrations/handlers/google_fit_handler/token.json', 'w') as token:
                token.write(creds.to_json())
        self.api = build('fitness', 'v1', credentials=creds)
        
        self.is_connected = True
        return self.api

    def check_connection(self) -> StatusResponse:
        response = StatusResponse(False)

        try:
            api = self.connect()
            response.success = True

        except Exception as e:
            log.logger.error(f'Error connecting to Google Fit API: {e}!')
            response.error_message = e

        self.is_connected = response.success
        return response

    def retrieve_data(self, service, startTimeMillis, endTimeMillis, dataSourceId) -> dict:
        try:
            return service.users().dataset().aggregate(userId="me", body={
                "aggregateBy": [{
                    "dataTypeName": "com.google.step_count.delta",
                    "dataSourceId": dataSourceId
                }],
                "bucketByTime": {"durationMillis": 86400000},
                "startTimeMillis": startTimeMillis,
                "endTimeMillis": endTimeMillis
            }).execute()
        except HttpError:
            raise HttpError


    def native_query(self, query: str = None) -> Response:
        """Receive raw query and act upon it somehow.
        Args:
            query (Any): query in native format (str for sql databases,
            dict for mongo, api's json etc)
        Returns:
            HandlerResponse
        """
        ast = parse_sql(query, dialect='mindsdb')
        return self.query(ast)
    
    def get_steps(self, start_time_millis, end_time_millis) -> pd.DataFrame:
        steps = {}
        steps_data = self.retrieve_data(self.api, start_time_millis, end_time_millis, "derived:com.google.step_count.delta:com.google.android.gms:estimated_steps")
        for daily_step_data in steps_data['bucket']:
            local_date = datetime.fromtimestamp(int(daily_step_data['startTimeMillis']) / 1000,
                                            tz=pytz.timezone(str(get_localzone())))
            local_date_str = local_date.strftime(DATE_FORMAT)

            data_point = daily_step_data['dataset'][0]['point']
            if data_point:
                count = data_point[0]['value'][0]['intVal']
                data_source_id = data_point[0]['originDataSourceId']
                steps[local_date_str] = {'steps': count, 'originDataSourceId': data_source_id}
        ret = pd.DataFrame.from_dict(steps)
        ret = ret.T
        ret = ret.drop('originDataSourceId', axis=1)
        ret = ret.reset_index(drop=False)
        return ret
    
    def call_google_fit_api(self, method_name:str = None, params:dict = None) -> pd.DataFrame:
        """Receive query as AST (abstract syntax tree) and act upon it somehow.
        Args:
            query (ASTNode): sql query represented as AST. May be any kind
                of query: SELECT, INSERT, DELETE, etc
        Returns:
            DataFrame
        """
        self.connect()
        if method_name == 'get_steps':
            val = self.get_steps(params['start_time'], params['end_time'])
            return val
        raise NotImplementedError('Method name {} not supported by Google Fit Handler'.format(method_name))


__title__ = 'MindsDB DynamoDB handler'
__package_name__ = 'mindsdb_dynamodb_handler'
__version__ = '0.0.1'
__description__ = "MindsDB handler for DynamoDB"
__author__ = 'Minura Punchihewa'
__github__ = 'https://github.com/mindsdb/mindsdb'
__pypi__ = 'https://pypi.org/project/mindsdb/'
__license__ = 'GPL-3.0'
__copyright__ = 'Copyright 2022- mindsdb'



from mindsdb.integrations.libs.const import HANDLER_TYPE

from .__about__ import __description__ as description
from .__about__ import __version__ as version

try:
    from .writer_handler import WriterHandler as Handler

    import_error = None
except Exception as e:
    Handler = None
    import_error = e

title = "Writer"
name = "writer"
type = HANDLER_TYPE.ML
permanent = False


__all__ = [
    "Handler",
    "version",
    "name",
    "type",
    "title",
    "description",
    "import_error",
]



import os
import time
import unittest
import tempfile

temp_dir = tempfile.mkdtemp(dir='/tmp/', prefix='lightwood_handler_test_')
os.environ['MINDSDB_STORAGE_DIR'] = os.environ.get('MINDSDB_STORAGE_DIR', temp_dir)
os.environ['MINDSDB_DB_CON'] = 'sqlite:///' + os.path.join(os.environ['MINDSDB_STORAGE_DIR'], 'mindsdb.sqlite3.db') + '?check_same_thread=False&timeout=30'

from mindsdb.migrations import migrate
migrate.migrate_to_head()

from mindsdb.utilities.config import Config
from mindsdb.integrations.utilities.test_utils import HandlerControllerMock, PG_HANDLER_NAME, PG_CONNECTION_DATA
from mindsdb.interfaces.database.integrations import integration_controller
from mindsdb.integrations.handlers.lightwood_handler.lightwood_handler.lightwood_handler import LightwoodHandler
# from mindsdb.integrations.handlers.lightwood_handler.lightwood_handler.utils import load_predictor
from mindsdb.interfaces.storage.fs import FsStore
from mindsdb.interfaces.model.model_controller import ModelController
from mindsdb.integrations.libs.response import RESPONSE_TYPE
import mindsdb.interfaces.storage.db as db


# TODO: drop all models and tables when closing tests
class LightwoodHandlerTest(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        # region create permanent integrations
        for integration_name in ['files', 'lightwood']:
            integration_record = db.Integration(
                name=integration_name,
                data={},
                engine=integration_name,
                company_id=None
            )
            db.session.add(integration_record)
            db.session.commit()
        integration_record = db.Integration(
            name=PG_HANDLER_NAME,
            data=PG_CONNECTION_DATA,
            engine='postgres',
            company_id=None
        )
        db.session.add(integration_record)
        db.session.commit()
        # endregion

        handler_controller = integration_controller

        cls.handler = LightwoodHandler(
            'lightwood',
            handler_controller=handler_controller,
            fs_store=FsStore(),
            model_controller=ModelController()
        )
        cls.config = Config()

        cls.target_1 = 'rental_price'
        cls.data_table_1 = 'demo_data.home_rentals'
        cls.test_model_1 = 'test_lightwood_home_rentals'
        cls.test_model_1b = 'test_lightwood_home_rentals_custom'

        cls.target_2 = 'Traffic'
        cls.data_table_2 = 'demo_data.house_sales'
        cls.test_model_2 = 'test_lightwood_house_sales'

    def test_00_check_connection(self):
        conn = self.handler.check_connection()
        assert conn.success

    def test_01_drop_predictor(self):
        if self.test_model_1 not in self.handler.get_tables().data_frame.values:
            # TODO: seems redundant because of test_02
            query = f"""
                CREATE PREDICTOR {self.test_model_1}
                FROM {PG_HANDLER_NAME} (SELECT * FROM {self.data_table_1} limit 50)
                PREDICT rental_price
            """
            self.handler.native_query(query)
        response = self.handler.native_query(f"DROP PREDICTOR {self.test_model_1}")
        self.assertTrue(response.type == RESPONSE_TYPE.OK)

    def test_02_train_predictor(self):
        query = f"""
            CREATE PREDICTOR {self.test_model_1}
            FROM {PG_HANDLER_NAME} (SELECT * FROM {self.data_table_1} limit 50)
            PREDICT rental_price
        """
        response = self.handler.native_query(query)
        time.sleep(5)
        self.assertTrue(response.type == RESPONSE_TYPE.OK)

    def test_03_retrain_predictor(self):
        query = f"RETRAIN {self.test_model_1}"
        response = self.handler.native_query(query)
        self.assertTrue(response.type == RESPONSE_TYPE.OK)

    def test_04_query_predictor_single_where_condition(self):
        time.sleep(120) # TODO 
        query = f"""
            SELECT target
            from {self.test_model_1}
            WHERE sqft=100
        """
        response = self.handler.native_query(query)
        self.assertTrue(response.type == RESPONSE_TYPE.TABLE)
        self.assertTrue(len(response.data_frame) == 1)
        self.assertTrue(response.data_frame['sqft'][0] == 100)
        self.assertTrue(response.data_frame['rental_price'][0] is not None)

    def test_05_query_predictor_multi_where_condition(self):
        query = f"""
            SELECT target
            from {self.test_model_1}
            WHERE sqft=100
                  AND number_of_rooms=2
                  AND number_of_bathrooms=1
        """
        response = self.handler.native_query(query)
        self.assertTrue(response.type == RESPONSE_TYPE.TABLE)
        self.assertTrue(len(response.data_frame) == 1)
        self.assertTrue(response.data_frame['number_of_rooms'][0] == 2)
        self.assertTrue(response.data_frame['number_of_bathrooms'][0] == 1)

    def test_06_train_predictor_custom_jsonai(self):
        # TODO: turn this into a decorator?
        if self.test_model_1b in self.handler.get_tables().data_frame.values:  # TODO this accesor feels weird, maybe rethink output format?
            self.handler.native_query(f"DROP PREDICTOR {self.test_model_1b}")

        using_str = 'model.args={"submodels": [{"module": "LightGBM", "args": {"stop_after": 12, "fit_on_dev": true}}]}'
        query = f"""
            CREATE PREDICTOR {self.test_model_1b}
            FROM {PG_HANDLER_NAME} (SELECT * FROM {self.data_table_1} limit 50)
            PREDICT rental_price
            USING {using_str}
        """
        response = self.handler.native_query(query)
        self.assertTrue(response.type == RESPONSE_TYPE.OK)
        # TODO assert
        # m = load_predictor(self.handler.storage.get('models')[self.test_model_1b], self.test_model_1b)
        # assert len(m.ensemble.mixers) == 1
        # assert type(m.ensemble.mixers[0]).__name__ == 'LightGBM'

    def test_07_list_tables(self):
        response = self.handler.get_tables()
        self.assertTrue(response.type == RESPONSE_TYPE.TABLE)

    def test_08_get_columns(self):
        response = self.handler.get_columns(f'{self.test_model_1}')
        self.assertTrue(response.type == RESPONSE_TYPE.TABLE)

    # TODO
    # def test_09_join_predictor_into_table(self):
    #     into_table = 'test_join_into_lw'
    #     query = f"SELECT tb.{self.target_1} as predicted, ta.{self.target_1} as truth, ta.sqft from {PG_HANDLER_NAME}.{self.data_table_1} AS ta JOIN {self.test_model_1} AS tb LIMIT 10"
    #     parsed = self.handler.parser(query, dialect=self.handler.dialect)
    #     predicted = self.handler.join(parsed, self.data_handler, into=into_table)

    #     # checks whether `into` kwarg does insert into the table or not
    #     q = f"SELECT * FROM {into_table}"
    #     qp = self.handler.parser(q, dialect='mysql')
    #     assert len(self.data_handler.query(qp).data_frame) > 0

    def test_10_train_ts_predictor_multigby_hor4(self):
        # TODO: handle cap/uncapped column name returned from data handler? Had to rename 'MA' -> 'ma' for test to pass
        query = f"""
            CREATE PREDICTOR {self.test_model_2}
            FROM {PG_HANDLER_NAME} (SELECT * FROM {self.data_table_2})
            PREDICT ma
            ORDER BY saledate
            GROUP BY bedrooms, type
            WINDOW 8
            HORIZON 4
        """
        if self.test_model_2 not in self.handler.get_tables().data_frame.values:
            response = self.handler.native_query(query)
        else:
            self.handler.native_query(f"DROP PREDICTOR {self.test_model_2}")
            response = self.handler.native_query(query)

        self.assertTrue(response.type == RESPONSE_TYPE.OK)

        # TODO: reactivate and add to the rest of the TS tests once cache is back on
        # p = self.handler.storage.get('models')
        # m = load_predictor(p[self.test_model_2], self.test_model_2)
        # assert m.problem_definition.timeseries_settings.is_timeseries

    def test_12_train_ts_predictor_multigby_hor1(self):
        query = f"""
            CREATE PREDICTOR {self.test_model_2}
            FROM {PG_HANDLER_NAME} (SELECT * FROM {self.data_table_2})
            PREDICT ma
            ORDER BY saledate
            GROUP BY bedrooms, type
            WINDOW 8
            HORIZON 1
        """
        if self.test_model_2 not in self.handler.get_tables().data_frame.values:
            self.handler.native_query(query)
        else:
            self.handler.native_query(f"DROP PREDICTOR {self.test_model_2}")
            self.handler.native_query(query)

    def test_13_train_ts_predictor_no_gby_hor1(self):
        query = f"""
            CREATE PREDICTOR {self.test_model_2}
            FROM {PG_HANDLER_NAME} (SELECT * FROM {self.data_table_2})
            PREDICT ma
            ORDER BY saledate
            WINDOW 8
            HORIZON 1
        """
        if self.test_model_2 not in self.handler.get_tables().data_frame.values:
            self.handler.native_query(query)
        else:
            self.handler.native_query(f"DROP PREDICTOR {self.test_model_2}")
            self.handler.native_query(query)

    def test_14_train_ts_predictor_no_gby_hor4(self):
        query = f"""
            CREATE PREDICTOR {self.test_model_2}
            FROM {PG_HANDLER_NAME} (SELECT * FROM {self.data_table_2})
            PREDICT ma
            ORDER BY saledate
            WINDOW 8
            HORIZON 4
        """
        if self.test_model_2 not in self.handler.get_tables().data_frame.values:
            self.handler.native_query(query)
        else:
            self.handler.native_query(f"DROP PREDICTOR {self.test_model_2}")
            self.handler.native_query(query)

    # TODO
    # def test_15_join_predictor_ts_into(self):
    #     query = f"""
    #         SELECT m.saledate as date,
    #                m.ma as forecast
    #         FROM mindsdb.{self.test_model_2} m JOIN {PG_HANDLER_NAME}.demo_data.house_sales t
    #         WHERE t.saledate > LATEST
    #               AND t.type = 'house'
    #               AND t.bedrooms = 2
    #         LIMIT 10
    #     """
    #     response = self.handler.native_query(query)
    #     self.assertTrue(response.type == RESPONSE_TYPE.TABLE)

    # def test_16_join_predictor_ts_model_left(self):
    #     # TODO: is this one needed?
    #     target = 'Traffic'
    #     oby = 'T'
    #     query = f"SELECT tb.{target} as predicted, ta.{target} as truth, ta.{oby} from mindsdb.{self.test_tsmodel_name_1} AS tb JOIN {self.sql_handler_name}.{self.data_table_name_2} AS ta ON 1=1 WHERE ta.{oby} > LATEST LIMIT 10"
    #     parsed = self.handler.parser(query, dialect=self.handler.dialect)
    #     predicted = self.handler.join(parsed, self.data_handler)  # , into=self.model_2_into_table) # TODO: restore when we add support for SQLite and other handlers for `into`


if __name__ == "__main__":
    unittest.main(failfast=True)


from mindsdb.api.mysql.mysql_proxy.controllers import SessionController
from mindsdb.api.mysql.mysql_proxy.executor import Executor


class SqlServerStub:
    """This class is just an emulation of Server object,
    used by Executor.
    In 'monolithic' mode of MindsDB work the Executor takes
    some information from the sql server which. Here we emulate
    this object."""

    def __init__(self, **kwargs):
        for arg in kwargs:
            setattr(self, arg, kwargs[arg])


def run_sql_command(request_env, ast_query):
    sql_session = SessionController()
    sql_session.database = request_env.get('database', 'mindsdb')
    sqlserver = SqlServerStub(connection_id=-1)

    executor = Executor(sql_session, sqlserver)
    executor.binary_query_execute(ast_query)
    if executor.error_code:
        raise Exception(executor.error_message)

    if not executor.columns:
        # return no data
        return []

    column_names = [
        c.name is c.alias is None or c.alias
        for c in executor.columns
    ]

    data = []
    for row in executor.data:
        data.append(dict(zip(column_names, row)))

    return data



import gc
import sys
import json
import dill

from lightwood.api.high_level import _module_from_code


def unpack_jsonai_old_args(json_ai_override):
    while '.' in str(list(json_ai_override.keys())):
        for k in list(json_ai_override.keys()):
            if '.' in k:
                nks = k.split('.')
                obj = json_ai_override
                for nk in nks[:-1]:
                    if nk not in obj:
                        obj[nk] = {}
                    obj = obj[nk]
                obj[nks[-1]] = json_ai_override[k]
                del json_ai_override[k]


def load_predictor(predictor_dict, name):
    try:
        return dill.loads(predictor_dict['predictor'])
    except Exception as e:
        module_name = str(e).lstrip("No module named '").split("'")[0]

        try:
            del sys.modules[module_name]
        except Exception:
            pass

        gc.collect()
        _module_from_code(predictor_dict['code'], module_name)
        return dill.loads(predictor_dict['predictor'])


def rep_recur(org: dict, ovr: dict):
    for k in ovr:
        if k in org:
            if isinstance(org[k], dict) and isinstance(ovr[k], dict):
                rep_recur(org[k], ovr[k])
            else:
                org[k] = ovr[k]
        else:
            org[k] = ovr[k]


def brack_to_mod(ovr):
    if not isinstance(ovr, dict):
        if isinstance(ovr, list):
            for i in range(len(ovr)):
                ovr[i] = brack_to_mod(ovr[i])
        elif isinstance(ovr, str):
            if '(' in ovr and ')' in ovr:
                mod = ovr.split('(')[0]
                args = {}
                if '()' not in ovr:
                    for str_pair in ovr.split('(')[1].split(')')[0].split(','):
                        k = str_pair.split('=')[0].strip(' ')
                        v = str_pair.split('=')[1].strip(' ')
                        args[k] = v

                ovr = {
                    'module': mod,
                    'args': args
                }
            elif '{' in ovr and '}' in ovr:
                try:
                    ovr = json.loads(ovr)
                except Exception:
                    pass
        return ovr
    else:
        for k in ovr.keys():
            ovr[k] = brack_to_mod(ovr[k])

    return ovr



import os
import requests
import tempfile
import shutil
from pathlib import Path
from zipfile import ZipFile

from mindsdb.utilities.config import Config
from mindsdb.utilities.log import get_log


def download_gui(destignation, version):
    if isinstance(destignation, str):
        destignation = Path(destignation)
    logger = get_log('http')
    dist_zip_path = str(destignation.joinpath('dist.zip'))
    bucket = "https://mindsdb-web-builds.s3.amazonaws.com/"

    resources = [{
        'url': bucket + 'dist-V' + version + '.zip',
        'path': dist_zip_path
    }]

    def get_resources(resource):
        response = requests.get(resource['url'])
        if response.status_code != requests.status_codes.codes.ok:
            raise Exception(f"Error {response.status_code} GET {resource['url']}")
        open(resource['path'], 'wb').write(response.content)
    try:
        for r in resources:
            get_resources(r)
    except Exception as e:
        logger.error(f'Error during downloading files from s3: {e}')
        return False

    static_folder = destignation
    static_folder.mkdir(mode=0o777, exist_ok=True, parents=True)
    ZipFile(dist_zip_path).extractall(static_folder)

    if static_folder.joinpath('dist').is_dir():
        shutil.move(str(destignation.joinpath('dist').joinpath('index.html')), static_folder)
        shutil.move(str(destignation.joinpath('dist').joinpath('assets')), static_folder)
        shutil.rmtree(destignation.joinpath('dist'))

    os.remove(dist_zip_path)

    version_txt_path = destignation.joinpath('version.txt')
    with open(version_txt_path, 'wt') as f:
        f.write(version)

    return True

    '''
    # to make downloading faster download each resource in a separate thread
    with concurrent.futures.ThreadPoolExecutor(max_workers=5) as executor:
        future_to_url = {executor.submit(get_resources, r): r for r in resources}
        for future in concurrent.futures.as_completed(future_to_url):
            res = future.result()
            if res is not None:
                raise res
    '''


def update_static(gui_version_lv):
    ''' Update Scout files basing on compatible-config.json content.
        Files will be downloaded and updated if new version of GUI > current.
        Current GUI version stored in static/version.txt.
    '''
    config = Config()
    logger = get_log('http')
    static_path = Path(config['paths']['static'])

    logger.info(f'New version of GUI available ({gui_version_lv.vstring}). Downloading...')

    temp_dir = tempfile.mkdtemp(prefix='mindsdb_gui_files_')
    success = download_gui(temp_dir, gui_version_lv.vstring)
    if success is False:
        shutil.rmtree(temp_dir)
        return False

    temp_dir_for_rm = tempfile.mkdtemp(prefix='mindsdb_gui_files_')
    shutil.rmtree(temp_dir_for_rm)
    shutil.copytree(str(static_path), temp_dir_for_rm)
    shutil.rmtree(str(static_path))
    shutil.copytree(temp_dir, str(static_path))
    shutil.rmtree(temp_dir_for_rm)

    logger.info(f'GUI version updated to {gui_version_lv.vstring}')
    return True



import json
import pickle

import grpc
from mindsdb.microservices_grpc.ml import ml_pb2_grpc
from mindsdb.microservices_grpc.ml import ml_pb2, common_pb2

from mindsdb.integrations.libs.response import (
    HandlerResponse as Response,
    HandlerStatusResponse as StatusResponse,
)

# from mindsdb.integrations.libs.ml_exec_base import BaseMLEngineExec
from mindsdb.integrations.libs.handler_helpers import action_logger
from mindsdb.utilities.context import context as ctx
from mindsdb.utilities.log import get_log


logger = get_log(logger_name="main")


class MLClientGRPC:
    def __init__(self, host, port, **handler_params):
        self.host = host
        self.port = port
        self.handler_params = handler_params
        self.integration_id = handler_params.get("integration_id")
        self.predictor_id = handler_params.get("predictor_id")
        for key in ("integration_id", "predictor_id"):
            try:
                del self.handler_params[key]
            except Exception:
                pass
        # have to create a handler instance
        # because Executor accesses to some handler attributes
        # directly
        # self.handler = BaseMLEngineExec(**self.handler_params)
        # remove all 'object' params from dict before sending it to the serverside.
        # all of them will be created there
        self.channel = grpc.insecure_channel(f"{self.host}:{self.port}")
        self.stub = ml_pb2_grpc.MLServiceStub(self.channel)

    def __del__(self):
        if hasattr(self, "channel"):
            self.channel.close()

    @property
    def context(self):
        return ml_pb2.HandlerContextML(
            predictor_id=self.predictor_id,
            integration_id=self.integration_id,
            context=json.dumps(ctx.dump()),
            handler_params=json.dumps(self.handler_params),
        )

    @staticmethod
    def _to_status_response(response: common_pb2.StatusResponse):
        return StatusResponse(success=response.success,
                              error_message=response.error_message)

    @staticmethod
    def _to_response(response: common_pb2.Response):
        data = pickle.loads(response.data_frame)
        return Response(
            resp_type=response.type,
            data_frame=data,
            query=response.query,
            error_code=response.error_code,
            error_message=response.error_message,
        )

    @action_logger(logger)
    def predict(
        self,
        df,
        args
    ):
        request = ml_pb2.PredictCall(context=self.context, df=pickle.dumps(df), args=json.dumps(args))
        resp = self.stub.Predict(request)

        logger.info("%s.learn: returned error - %s, error_message - %s", self.__class__.__name__, resp.error_code, resp.error_message)
        if resp.error_code and resp.error_message:
            raise Exception(resp.error_message)

        return pickle.loads(resp.data_frame)

    @action_logger(logger)
    def create(
            self,
            target,
            df,
            args
    ):

        request = ml_pb2.CreateCall(context=self.context, target=target, df=pickle.dumps(df), args=json.dumps(args))
        resp = self.stub.Create(request)

        logger.info("%s.learn: success - %s", self.__class__.__name__, resp.success)
        if not resp.success:
            logger.error("%s.learn: returned error - %s", self.__class__.__name__, resp.error_message)
            raise Exception(resp.error_message)

    @action_logger(logger)
    def update(
        self,
        df,
        args
    ):

        request = ml_pb2.UpdateCall(context=self.context, df=pickle.dumps(df), args=json.dumps(args))
        resp = self.stub.Update(request)

        logger.info("%s.update: success - %s", self.__class__.__name__, resp.success)
        if not resp.success:
            logger.error("%s.update: returned error - %s", self.__class__.__name__, resp.error_message)
            raise Exception(resp.error_message)

    @action_logger(logger)
    def close(self):
        pass



import time
import importlib
from threading import Thread

import requests

from mindsdb.utilities.log import get_log
logger = get_log(logger_name="main")


def action_logger(logger, loglevel="info"):
    def decorator(func):
        def wrapper(*args, **kwargs):
            log_func = getattr(logger, loglevel)
            instance = args[0]
            log_func("%s.%s: calling with args - %s, kwargs - %s", instance.__class__.__name__, func.__name__, args[1:], kwargs)
            res = func(*args, **kwargs)
            log_func("%s.%s: returning - %s", instance.__class__.__name__, func.__name__, res)
            return res
        return wrapper
    return decorator


def get_handler(_type):
    _type = _type.lower()
    # a crutch to fix bug in handler naming convention
    if _type == "files":
        _type = "file"
    handler_folder_name = _type + "_handler"
    logger.debug("get_handler: handler_folder - %s", handler_folder_name)

    try:
        handler_module = importlib.import_module(f'mindsdb.integrations.handlers.{handler_folder_name}')
        logger.debug("get_handler: handler module - %s", handler_module)
        handler = handler_module.Handler
        if handler is None:
            logger.error("get_handler: import error - %s", handler_module.import_error)
        logger.debug("get_handler: found handler - %s", handler)
        return handler
    except Exception as e:
        raise e


def registry(url, data, interval):
    def worker():
        while True:
            try:
                requests.post(url, json=data, headers={"Content-Type": "application/json"})
            except Exception as e:
                logger.error("handler_register: unable to register the service - %s", e)
            time.sleep(interval)

    _th_worker = Thread(target=worker)
    _th_worker.start()
    logger.info("handler_register: registering service. url - %s, data - %s, interval - %s", url, data, interval)


def discover_services(url):
    res = None
    try:
        res = requests.get(url, headers={"Content-Type": "application/json"})
        res = res.json()
    except Exception as e:
        logger.error("service discover: unable to get handlers metadata. url - %s, error - %s", url, e)
        res = {}
    logger.info("discover_services: service response - %s", res)
    return res



import unittest
from mindsdb.integrations.handlers.timescaledb_handler.timescaledb_handler import TimeScaleDBHandler
from mindsdb.api.mysql.mysql_proxy.libs.constants.response_type import RESPONSE_TYPE


class TimeScaleDBHandlerTest(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.kwargs = {
            "host": "localhost",
            "port": 36806,
            "user": "USERNAME",
            "password": "P455W0rd",
            "database": "tsdb"
        }
        cls.handler = TimeScaleDBHandler('test_timescaledb_handler', connection_data=cls.kwargs)

    def test_0_connect(self):
        self.handler.connect()

    def test_1_drop_table(self):
        res = self.handler.query("DROP TABLE IF EXISTS LOVE;")
        assert res.type is not RESPONSE_TYPE.ERROR

    def test_2_create_table(self):
        res = self.handler.query("CREATE TABLE IF NOT EXISTS LOVE (LOVER varchar(20));")
        assert res.type is not RESPONSE_TYPE.ERROR
    
    def test_3_insert(self):
        res = self.handler.query("INSERT INTO LOVE VALUES('Hari');")
        assert res.type is not RESPONSE_TYPE.ERROR
    
    def test_4_get_tables(self):
        tables = self.handler.get_tables()
        assert tables.type is RESPONSE_TYPE.TABLE

    def test_5_select_query(self):
        query = "SELECT * FROM LOVE;"
        result = self.handler.native_query(query)
        assert result.type is RESPONSE_TYPE.TABLE

    def test_5_check_connection(self):
        self.handler.check_connection()

        
if __name__ == '__main__':
    unittest.main()


__title__ = "MindsDB PayPal handler"
__package_name__ = "mindsdb_paypal_handler"
__version__ = "0.0.1"
__description__ = "MindsDB handler for PayPal"
__author__ = "Minura Punchihewa"
__github__ = "https://github.com/mindsdb/mindsdb"
__pypi__ = "https://pypi.org/project/mindsdb/"
__license__ = "GPL-3.0"
__copyright__ = "Copyright 2023 - mindsdb"



import json
import pickle
import traceback
from concurrent import futures

import grpc
from mindsdb.microservices_grpc.ml import ml_pb2_grpc
from mindsdb.microservices_grpc.ml import ml_pb2, common_pb2

from mindsdb.interfaces.storage.fs import (
    FileStorage,
    FileStorageFactory,
    RESOURCE_GROUP,
)
from mindsdb.interfaces.storage.model_fs import ModelStorage, HandlerStorage

from mindsdb.integrations.libs.response import (
    RESPONSE_TYPE,
)
from mindsdb.interfaces.database.integrations import integration_controller
from mindsdb.integrations.libs.handler_helpers import get_handler
from mindsdb.utilities.context import context as ctx
from mindsdb.utilities.log import get_log


logger = get_log(logger_name="main")


class MLServiceServicer(ml_pb2_grpc.MLServiceServicer):

    def __init__(self):
        logger.info(
            "%s.__init__: ", self.__class__.__name__
        )

    def run(self, **kwargs):
        host = kwargs.get("host", "127.0.0.1")
        port = kwargs.get("port", 50052)
        addr = f"{host}:{port}"
        server = grpc.server(futures.ThreadPoolExecutor(max_workers=10))
        ml_pb2_grpc.add_MLServiceServicer_to_server(
            MLServiceServicer(), server)
        server.add_insecure_port(addr)
        server.start()
        server.wait_for_termination()

    def _get_handler_controller(self):
        return integration_controller

    def _get_file_storage(self, integration_id):
        fs_store = FileStorage(
            resource_group=RESOURCE_GROUP.INTEGRATION,
            resource_id=integration_id,
            sync=True,
        )
        return fs_store

    def _get_storage_factory(self):
        return FileStorageFactory(resource_group=RESOURCE_GROUP.PREDICTOR, sync=True)

    def _get_handler(self, handler_context: ml_pb2.HandlerContextML):
        ctx.load(json.loads(handler_context.context))
        params = json.loads(handler_context.handler_params)
        logger.info(
            "%s._get_handler: create handler. params - %s",
            self.__class__.__name__,
            params,
        )
        integration_id = handler_context.integration_id
        predictor_id = handler_context.predictor_id
        _type = params.get("engine")
        del params["engine"]
        logger.info(
            "%s.get_handler: request handler of type - %s, context - %s",
            self.__class__.__name__,
            _type,
            params,
        )

        HandlerClass = get_handler(_type)
        handlerStorage = HandlerStorage(integration_id)
        modelStorage = ModelStorage(predictor_id)

        ml_handler = HandlerClass(
            engine_storage=handlerStorage,
            model_storage=modelStorage,
            **params
        )
        return ml_handler

    def Predict(self, request, context):
        result = None
        try:
            args = json.loads(request.args)
            df = pickle.loads(request.df)
            logger.info(
                "%s.Predict: args - %s",
                self.__class__.__name__,
                args,
            )

            handler = self._get_handler(request.context)
            predictions = handler.predict(df, args)
            handler.close()

            logger.info(
                "%s.Predict: got predictions - %s(type - %s)",
                self.__class__.__name__,
                predictions,
                type(predictions),
            )
            predictions = pickle.dumps(predictions)
            result = common_pb2.Response(
                type=RESPONSE_TYPE.OK,
                data_frame=predictions,
                query=0,
                error_code=0,
                error_message="",
            )

        except Exception:
            msg = traceback.format_exc()
            logger.error("%s.Predict: error - %s", self.__class__.__name__, msg)
            result = common_pb2.Response(
                type=RESPONSE_TYPE.ERROR,
                data_frame=None,
                query=0,
                error_code=1,
                error_message=msg,
            )
        return result

    def Create(self, request, context):
        result = None
        try:
            args = json.loads(request.args)
            target = request.target
            df = pickle.loads(request.df)
            logger.info(
                "%s.Create: target - %s df - %s, args - %s",
                self.__class__.__name__,
                target,
                df,
                args,
            )

            handler = self._get_handler(request.context)
            handler.create(target, df=df, args=args)

            result = common_pb2.StatusResponse(success=True, error_message="")

        except Exception:
            msg = traceback.format_exc()
            logger.error("%s.Create: error - %s", self.__class__.__name__, msg)
            result = common_pb2.StatusResponse(success=False, error_message=msg)

        return result

    def Update(self, request, context):

        result = None
        try:
            args = json.loads(request.args)
            df = pickle.loads(request.df)
            logger.info(
                "%s.Update: args - %s",
                self.__class__.__name__,
                args,
            )

            handler = self._get_handler(request.context)
            handler.update(df, args)

            result = common_pb2.StatusResponse(success=True, error_message="")

        except Exception:
            msg = traceback.format_exc()
            logger.error("%s.Update: error - %s", self.__class__.__name__, msg)
            result = common_pb2.StatusResponse(success=False, error_message=msg)

        return result


def serve():
    server = grpc.server(futures.ThreadPoolExecutor(max_workers=10))
    ml_pb2_grpc.add_MLServiceServicer_to_server(
        MLServiceServicer(), server)
    server.add_insecure_port('[::]:50052')
    print("staring rpc server on [::]:50052")
    logger.info("staring rpc server on [::]:50052")
    server.start()
    server.wait_for_termination()


if __name__ == '__main__':
    serve()



import unittest
from mindsdb.integrations.handlers.s3_handler.s3_handler import S3Handler
from mindsdb.api.mysql.mysql_proxy.libs.constants.response_type import RESPONSE_TYPE


class S3HandlerTest(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.kwargs = {
            "aws_access_key_id": "PCAQ2LJDOSWLNSQKOCPW",
            "aws_secret_access_key": "U/VjewPlNopsDmmwItl34r2neyC6WhZpUiip57i",
            "region_name": "us-east-1",
            "bucket": "mindsdb-bucket",
            "key": "iris.csv",
            "input_serialization": "{'CSV': {'FileHeaderInfo': 'NONE'}}"
        }
        cls.handler = S3Handler('test_s3_handler', cls.kwargs)

    def test_0_check_connection(self):
        assert self.handler.check_connection()

    def test_1_native_query_select(self):
        query = "SELECT * FROM S3Object"
        result = self.handler.native_query(query)
        assert result.type is RESPONSE_TYPE.TABLE

    def test_2_get_tables(self):
        tables = self.handler.get_tables()
        assert tables.type is not RESPONSE_TYPE.ERROR

    def test_3_get_columns(self):
        columns = self.handler.get_columns()
        assert columns.type is not RESPONSE_TYPE.ERROR


if __name__ == '__main__':
    unittest.main()



from pandas import DataFrame
from snowflake import connector
from snowflake.sqlalchemy import snowdialect

from mindsdb_sql.render.sqlalchemy_render import SqlalchemyRender
from mindsdb_sql.parser.ast.base import ASTNode

from mindsdb.integrations.libs.base import DatabaseHandler
from mindsdb.integrations.libs.response import (
    HandlerStatusResponse as StatusResponse,
    HandlerResponse as Response,
    RESPONSE_TYPE
)
from mindsdb.utilities import log


class SnowflakeHandler(DatabaseHandler):
    """
    This handler handles connection and execution of the Snowflake statements.
    """

    name = 'snowflake'

    def __init__(self, name, **kwargs):
        super().__init__(name)
        self.connection_data = kwargs.get('connection_data')
        self.is_connected = False
        self.connection = None

    def connect(self):
        if self.is_connected is True:
            return self.connection
        con = connector.connect(
            host=self.connection_data['host'],
            user=self.connection_data['user'],
            password=self.connection_data['password'],
            account=self.connection_data['account'],
            warehouse=self.connection_data['warehouse'],
            database=self.connection_data['database'],
            schema=self.connection_data['schema'],
            protocol=self.connection_data['protocol'],
            port=self.connection_data['port'],
            application='MindsDB'
        )
        self.is_connected = True
        self.connection = con
        return self.connection

    def disconnect(self):
        if self.is_connected is False:
            return
        self.connection.close()
        self.is_connected = False
        return 

    def check_connection(self) -> StatusResponse:
        """
        Check the connection
        :return: success status and error message if error occurs
        """
        response = StatusResponse(False)
        need_to_close = self.is_connected is False
        try:
            connection = self.connect()
            with connection.cursor() as cur:
                cur.execute('select 1;')
            response.success = True
        except connector.errors.Error as e:
            log.logger.error(f'Error connecting to Snowflake {self.connection_data["database"]}, {e}!')
            response.error_message = e

        if response.success is True and need_to_close:
            self.disconnect()

        if response.success is False and self.is_connected is True:
            self.is_connected = False
        return response

    def native_query(self, query: str) -> Response:
        """
        Receive SQL query and runs it
        :param query: The SQL query to run in Snowflake
        :return: returns the records from the current recordset
        """
        need_to_close = self.is_connected is False
        connection = self.connect()
        from snowflake.connector import DictCursor
        with connection.cursor(DictCursor) as cur:
            try:
                cur.execute(query)
                result = cur.fetchall()
                if result:
                    response = Response(
                        RESPONSE_TYPE.TABLE,
                        DataFrame(
                            result,
                            columns=[x[0] for x in cur.description]
                        )
                    )
                else:
                    response = Response(RESPONSE_TYPE.OK)
            except Exception as e:
                log.logger.error(f'Error running query: {query} on {self.connection_data["database"]}!')
                response = Response(
                    RESPONSE_TYPE.ERROR,
                    error_message=str(e)
                )
        if need_to_close is True:
            self.disconnect()
        return response

    def get_tables(self) -> Response:
        """
        Get a list with all of the tabels in the current database or schema
        that the user has acces to
        """
        q = "SHOW TABLES;"
        result = self.native_query(q)
        result.data_frame = result.data_frame.rename(columns={'name': 'table_name'})
        return result

    def get_columns(self, table_name) -> Response:
        """
        List the columns in the tabels for which the user have access
        """
        q = f"SHOW COLUMNS IN TABLE {table_name};"
        result = self.native_query(q)
        return result

    def query(self, query: ASTNode) -> Response:
        """
        Retrieve the data from the SQL statement.
        """
        renderer = SqlalchemyRender(snowdialect.dialect)
        query_str = renderer.get_string(query, with_failback=True)
        return self.native_query(query_str)



import pandas as pd

from mindsdb_sql.parser import ast
from mindsdb_sql.parser.ast.base import ASTNode
from mindsdb_sql.planner.utils import query_traversal


def make_sql_session():
    from mindsdb.api.mysql.mysql_proxy.controllers.session_controller import SessionController

    sql_session = SessionController()
    sql_session.database = 'mindsdb'
    return sql_session


def conditions_to_filter(binary_op: ASTNode):
    conditions = extract_comparison_conditions(binary_op)

    filters = {}
    for op, arg1, arg2 in conditions:
        if op != '=':
            raise NotImplementedError
        filters[arg1] = arg2
    return filters


def extract_comparison_conditions(binary_op: ASTNode):
    '''Extracts all simple comparison conditions that must be true from an AST node.
    Does NOT support 'or' conditions.
    '''
    conditions = []

    def _extract_comparison_conditions(node: ASTNode, **kwargs):
        if isinstance(node, ast.BinaryOperation):
            op = node.op.lower()
            if op == 'and':
                # Want to separate individual conditions, not include 'and' as its own condition.
                return
            elif not isinstance(node.args[0], ast.Identifier):
                # Only support [identifier] =/</>/>=/<=/etc [constant] comparisons.
                raise NotImplementedError(f'Not implemented arg1: {node.args[0]}')

            if isinstance(node.args[1], ast.Constant):
                value = node.args[1].value
            elif isinstance(node.args[1], ast.Tuple):
                value = [i.value for i in node.args[1].items]
            else:
                raise NotImplementedError(f'Not implemented arg2: {node.args[1]}')

            conditions.append([op, node.args[0].parts[-1], value])

    query_traversal(binary_op, _extract_comparison_conditions)
    return conditions


def project_dataframe(df, targets, table_columns):
    columns = []
    for target in targets:
        if isinstance(target, ast.Star):
            columns = table_columns
            break
        elif isinstance(target, ast.Identifier):
            columns.append(target.parts[-1])
        else:
            raise NotImplementedError

    if len(df) == 0:
        df = pd.DataFrame([], columns=columns)
    else:
        # add absent columns
        for col in set(columns) & set(df.columns) ^ set(columns):
            df[col] = None

        # filter by columns
        df = df[columns]
    return df



import datetime as dt

from mindsdb_sql.parser.ast import Identifier, Select, Insert, BinaryOperation, Constant
from mindsdb.interfaces.storage import db

from mindsdb.integrations.libs.api_handler import APIChatHandler


class ChatBotTask:
    def __init__(self, session, database, project_name, model_name, params):

        self.session = session

        self.model_name = model_name
        self.db_handler = self.session.integration_controller.get_handler(database)
        self.project_datanode = self.session.datahub.get(project_name)

        # get chat handler info
        self.params = {}
        if isinstance(self.db_handler, APIChatHandler):
            self.params = self.db_handler.get_chat_config()

            # get bot username
            self.params['bot_username'] = self.db_handler.get_my_user_name()

        if params is not None:
            self.params.update(params)

        self.chats_prev = None

        # get model info
        model = self.session.model_controller.get_model(model_name, project_name=project_name)
        model_record = db.Predictor.query.get(model['id'])
        integration_record = db.Integration.query.get(model_record.integration_id)

        self.params['model'] = {
            'user_column': model_record.learn_args['using']['user_column'],
            'bot_column': model_record.learn_args['using']['assistant_column'],
            'output': model_record.to_predict[0],
            'engine': integration_record.engine,
        }

        self.params['back_db'] = {}
        back_db = self.params.get('backoffice_db')
        if back_db is not None:
            self.back_db = self.session.integration_controller.get_handler(back_db)

            if hasattr(self.back_db, 'back_office_config'):
                self.params['back_db']['config'] = self.back_db.back_office_config()

            if 'tools' in self.params['back_db']['config']:
                if self.params['model']['engine'] == 'langchain':
                    tools = [
                        {
                            'name': name,
                            'func': getattr(self.back_db, name),
                            'description': description
                        }
                        for name, description in self.params['back_db']['config']['tools'].items()
                    ]
                    self.params['back_db']['tools'] = tools

    def run(self):

        # get chat ids to react
        if self.params['polling']['type'] == 'message_count':
            chat_ids = self.get_chats_by_message_count()
        else:
            raise NotImplementedError

        for chat_id in chat_ids:
            self.answer_to_user(chat_id)

    def get_chats_by_message_count(self):

        p_params = self.params['polling']

        chat_ids = []

        id_col = p_params['chat_id_col']
        msgs_col = p_params['count_col']
        # get chats status info
        ast_query = Select(
            targets=[
                Identifier(id_col),
                Identifier(msgs_col)],
            from_table=Identifier(p_params['table'])
        )

        resp = self.db_handler.query(query=ast_query)
        if resp.data_frame is None:
            raise Exception()

        chats = {}
        for row in resp.data_frame.to_dict('records'):
            chat_id = row[id_col]
            msgs = row[msgs_col]

            chats[chat_id] = msgs

        if self.chats_prev is None:
            # first run
            self.chats_prev = chats
        else:
            # compare
            # for new keys
            for chat_id, count_msgs in chats.items():
                if self.chats_prev.get(chat_id) != count_msgs:
                    chat_ids.append(chat_id)

            self.chats_prev = chats
        return chat_ids

    def answer_to_user(self, chat_id):
        bot_username = self.params['bot_username']

        t_params = self.params['chat_table']

        text_col = t_params['text_col']
        username_col = t_params['username_col']

        ast_query = Select(
            targets=[Identifier(text_col),
                     Identifier(username_col)],
            from_table=Identifier(t_params['name']),
            where=BinaryOperation(
                op='=',
                args=[
                    Identifier(t_params['chat_id_col']),
                    Constant(chat_id)
                ]
            )
        )

        resp = self.db_handler.query(ast_query)
        if resp.data_frame is None:
            raise Exception()

        df = resp.data_frame

        # check first message:
        if len(df) == 0:
            return
        if df[username_col][0] == bot_username:
            # the last message is from bot
            return

        question_col = self.params['model']['user_column']
        answer_col = self.params['model']['bot_column']

        messages = []
        msgs = df.to_dict('records')
        # sort by time
        msgs.reverse()
        for row in msgs:
            text = row[text_col]

            if text is None or text.strip() == '':
                # skip empty rows
                continue

            if row[username_col] != bot_username:
                # create new message row
                messages.append({question_col: text, answer_col: None})
            else:
                if len(messages) == 0:
                    # add empty row
                    messages.append({question_col: None, answer_col: None})

                # update answer in previous column
                messages[-1][answer_col] = text

        model_output = self.apply_model(messages, chat_id)

        # send answer to user
        ast_query = Insert(
            table=Identifier(t_params['name']),
            columns=[t_params['chat_id_col'], t_params['text_col']],
            values=[
                [chat_id, model_output],
            ]
        )

        self.db_handler.query(ast_query)

    def apply_model(self, messages, chat_id):

        tools = None
        if 'tools' in self.params['back_db']:
            tools = self.params['back_db']['tools']

        context_list = [
            f"- Today's date is {dt.datetime.now().strftime('%Y-%m-%d')}. It must be used to understand the input date from string like 'tomorrow', 'today', 'yesterday'"
        ]
        context = '\n'.join(context_list)

        # call model
        predictions = self.project_datanode.predict(
            model_name=self.model_name,
            data=messages,
            params={'tools': tools, 'context': context, 'max_iterations': 10}
        )

        output_col = self.params['model']['output']
        model_output = predictions.iloc[-1][output_col]
        return model_output



class DataNode:
    type = 'meta'

    def __init__(self):
        pass

    def get_type(self):
        return self.type

    def get_tables(self):
        pass

    def has_table(self, tableName):
        pass

    def get_table_columns(self, tableName):
        pass

    def query(self, query=None, native_query=None, session=None):
        return []



import os
import pickle
import json

import grpc
from mindsdb.microservices_grpc.db import db_pb2_grpc
from mindsdb.microservices_grpc.db import db_pb2, common_pb2

from mindsdb.integrations.libs.response import (
    HandlerStatusResponse as StatusResponse,
    HandlerResponse as Response,
)
from mindsdb.integrations.libs.handler_helpers import action_logger
from mindsdb.utilities.context import context as ctx
from mindsdb.utilities.log import get_log


logger = get_log(logger_name="main")


class DBClientGRPC:

    def __init__(self, handler_type: str, **kwargs: dict):
        self.handler_type = handler_type
        self.handler_params = kwargs
        host = os.environ.get("MINDSDB_DB_SERVICE_HOST", None)
        port = os.environ.get("MINDSDB_DB_SERVICE_PORT", None)
        for a in ("fs_store", "file_storage"):
            if a in self.handler_params:
                del self.handler_params[a]

        # FileController is an object
        # so it is not a good idea to send it
        # to service side as parameter
        # will create a separate instace instead
        if self.handler_type == "files":
            del self.handler_params["file_controller"]
        self.channel = grpc.insecure_channel(f"{host}:{port}")
        self.stub = db_pb2_grpc.DBServiceStub(self.channel)

    def __del__(self):
        self.channel.close()

    @property
    def context(self):
        ctx_str = json.dumps(ctx.dump())
        return db_pb2.HandlerContext(handler_type=self.handler_type,
                                     handler_params=json.dumps(self.handler_params),
                                     context=ctx_str)

    @staticmethod
    def _to_status_response(response: common_pb2.StatusResponse):
        return StatusResponse(success=response.success,
                              error_message=response.error_message)

    @staticmethod
    def _to_response(response: common_pb2.Response):
        data = pickle.loads(response.data_frame)
        return Response(
            resp_type=response.type,
            data_frame=data,
            query=response.query,
            error_code=response.error_code,
            error_message=response.error_message,
        )

    @action_logger(logger)
    def connect(self):
        resp = self.stub.Connect(self.context)
        logger.info(
            "%s.connect: returns success - %s, error - %s",
            self.__class__.__name__,
            resp.success,
            resp.error_message
        )

        return self._to_status_response(resp)

    @action_logger(logger)
    def check_connection(self):
        resp = self.stub.CheckConnection(self.context)
        logger.info(
            "%s.check_connection: returns success - %s, error - %s",
            self.__class__.__name__,
            resp.success,
            resp.error_message
        )

        return self._to_status_response(resp)

    @action_logger(logger)
    def disconnect(self):
        resp = self.stub.Disconnect(self.context)
        logger.info(
            "%s.disconnect: returns success - %s, error - %s",
            self.__class__.__name__,
            resp.success,
            resp.error_message,
        )

        return self._to_status_response(resp)

    @action_logger(logger)
    def native_query(self, query):
        request = db_pb2.NativeQueryContext(context=self.context, query=query)
        resp = self.stub.NativeQuery(request)
        logger.info("%s.native_query: returned error - %s, error_message - %s", self.__class__.__name__, resp.error_code, resp.error_message)
        data = pickle.loads(resp.data_frame)
        logger.info("%s.native_query: returned data(type of %s) - %s", self.__class__.__name__, type(data), data)

        return self._to_response(resp)

    @action_logger(logger)
    def query(self, query):
        query = pickle.dumps(query)
        request = db_pb2.BinaryQueryContext(context=self.context, query=query)
        resp = self.stub.BinaryQuery(request)
        logger.info("%s.query: returned error - %s, error_message - %s", self.__class__.__name__, resp.error_code, resp.error_message)
        data = pickle.loads(resp.data_frame)
        logger.info("%s.query: returned data(type of %s) - %s", self.__class__.__name__, type(data), data)

        return self._to_response(resp)

    @action_logger(logger)
    def get_tables(self):
        resp = self.stub.GetTables(self.context)
        logger.info("%s.get_tables: returned error - %s, error_message - %s", self.__class__.__name__, resp.error_code, resp.error_message)
        data = pickle.loads(resp.data_frame)
        logger.info("%s.get_tables: returned data(type of %s) - %s", self.__class__.__name__, type(data), data)

        return self._to_response(resp)

    @action_logger(logger)
    def get_columns(self, table):
        request = db_pb2.ColumnsContext(context=self.context, table=table)
        resp = self.stub.GetColumns(request)
        logger.info("%s.get_columns: returned error - %s, error_message - %s", self.__class__.__name__, resp.error_code, resp.error_message)
        data = pickle.loads(resp.data_frame)
        logger.info("%s.get_columns: returned data(type of %s) - %s", self.__class__.__name__, type(data), data)

        return self._to_response(resp)



from typing import List

import pandas as pd

from mindsdb.integrations.libs.api_handler import APITable
from mindsdb.integrations.utilities.sql_utils import conditions_to_filter, project_dataframe
from mindsdb_sql.parser import ast


def message_to_dataframe_row(message: dict):
    message['id'] = message['_id']
    message['room_id'] = message['rid']
    message['text'] = message['msg']
    message['sent_at'] = message['ts']

    if 'u' in message:
        if 'username' in message['u']:
            message['username'] = message['u']['username']
        if 'name' in message['u']:
            message['name'] = message['u']['name']
    if 'bot' in message and 'i' in message['bot']:
        message['bot_id'] = message['bot']['i']
    return message


class ChannelsTable(APITable):
    def select(self, query: ast.Select) -> pd.DataFrame:
        message_data = self.handler.call_api('channels_list')
        df = pd.DataFrame(message_data['channels'])
        df = project_dataframe(df, query.targets, self.get_columns())

        return df

    def get_columns(self):
        """Gets all columns to be returned in pandas DataFrame responses"""
        return [
            '_id',
            'name',
            'usersCount',
            'msgs',
        ]


class ChannelMessagesTable(APITable):
    """Manages SELECT and INSERT operations for Rocket Chat messages."""

    def select(self, query: ast.Select) -> pd.DataFrame:
        """Selects message data from the Rocket Chat API and returns it as a pandas DataFrame.

        Returns dataframe representing the Rocket Chat API results.

        Args:
            query (ast.Select): Given SQL SELECT query
        """
        filters = conditions_to_filter(query.where)

        if 'room_id' not in filters:
            raise NotImplementedError()

        params = {}

        if query.limit:
            params['count'] = query.limit

        # See Channel Messages endpoint:
        # https://developer.rocket.chat/reference/api/rest-api/endpoints/core-endpoints/channels-endpoints/messages
        message_data = self.handler.call_api('channels_history', filters['room_id'], **params)

        # Only return the columns we need to.
        message_rows = [message_to_dataframe_row(m) for m in message_data['messages']]
        df = pd.DataFrame(message_rows)

        df = project_dataframe(df, query.targets, self.get_columns())

        return df

    def insert(self, query: ast.Insert):
        """Posts a message using the Rocket Chat API.

        Args:
            query (ast.Insert): Given SQL INSERT query
        """
        # See Post Message endpoint:
        # https://developer.rocket.chat/reference/api/rest-api/endpoints/core-endpoints/chat-endpoints/postmessage
        column_names = [col.name for col in query.columns]
        for insert_row in query.values:
            insert_params = dict(zip(column_names, insert_row))

            self.handler.call_api('chat_post_message', **insert_params)

    def get_columns(self):
        """Gets all columns to be returned in pandas DataFrame responses"""
        return [
            'id',
            'room_id',
            'bot_id',
            'text',
            'username',
            'name',
            'sent_at'
        ]


class DirectsTable(APITable):
    def select(self, query: ast.Select) -> pd.DataFrame:
        message_data = self.handler.call_api('im_list')
        df = pd.DataFrame(message_data['ims'])
        df = project_dataframe(df, query.targets, self.get_columns())

        return df

    def insert(self, query: ast.Insert):
        column_names = [col.name for col in query.columns]
        for insert_row in query.values:
            insert_params = dict(zip(column_names, insert_row))

            self.handler.call_api('im_create', **insert_params)

    def get_columns(self):
        """Gets all columns to be returned in pandas DataFrame responses"""
        return [
            '_id',
            'usernames',
            'usersCount',
            'msgs',
        ]


class DirectMessagesTable(APITable):

    def select(self, query: ast.Select) -> pd.DataFrame:

        filters = conditions_to_filter(query.where)

        if 'room_id' not in filters:
            raise NotImplementedError()

        params = {}

        if query.limit:
            params['count'] = query.limit

        message_data = self.handler.call_api('im_history', filters['room_id'], **params)

        message_rows = [message_to_dataframe_row(m) for m in message_data['messages']]
        df = pd.DataFrame(message_rows)

        df = project_dataframe(df, query.targets, self.get_columns())

        return df

    def insert(self, query: ast.Insert):

        column_names = [col.name for col in query.columns]
        for insert_row in query.values:
            insert_params = dict(zip(column_names, insert_row))

            if 'username' in insert_params:
                # resolve username
                resp = self.handler.call_api('users_info', insert_params['username'])
                if 'user' in resp:
                    insert_params['room_id'] = resp['user']['_id']
                    del insert_params['username']
                else:
                    raise ValueError(f'User not found: {insert_params["username"]}')

            self.handler.call_api('chat_post_message', **insert_params)

    def get_columns(self):
        """Gets all columns to be returned in pandas DataFrame responses"""
        return [
            'id',
            'room_id',
            'bot_id',
            'text',
            'username',
            'name',
            'sent_at'
        ]


class UsersTable(APITable):
    def select(self, query: ast.Select) -> pd.DataFrame:
        message_data = self.handler.call_api('users_list')
        df = pd.DataFrame(message_data['users'])
        df = project_dataframe(df, query.targets, self.get_columns())

        return df

    def get_columns(self):
        """Gets all columns to be returned in pandas DataFrame responses"""
        return [
            '_id',
            'username',
            'name',
            'status',
            'active',
            'type',
        ]




from collections import defaultdict
from datetime import datetime
from pathlib import Path

import pandas as pd
from langchain.chains import RetrievalQA
from langchain.llms import Writer

from mindsdb.utilities.log import get_log

from .settings import DEFAULT_EMBEDDINGS_MODEL, ModelParameters, get_retriever

logger = get_log(logger_name=__name__)


class QuestionAnswerer:
    def __init__(self, args: dict, model_parameters: ModelParameters):
        self.output_data = defaultdict(list)

        self.embeddings_model_name = args.get(
            "embeddings_model_name", DEFAULT_EMBEDDINGS_MODEL
        )
        self.persist_directory = args["chromadb_storage_path"]

        llm = Writer(**model_parameters.dict())
        retriever = get_retriever(
            embeddings_model_name=self.embeddings_model_name,
            persist_directory=self.persist_directory,
        )

        self.qa = RetrievalQA.from_chain_type(
            llm=llm,
            chain_type="stuff",
            retriever=retriever,
            return_source_documents=True,
            callbacks=model_parameters.callbacks,
        )

    @property
    def results_df(self):
        return pd.DataFrame(self.output_data)

    def query(self, input_query: str):
        logger.debug(f"Querying: {input_query}")

        res = self.qa(input_query)

        self.output_data["question"].append(input_query)
        self.output_data["answer"].append(res["result"])

        sources = defaultdict(list)

        for idx, document in enumerate(res["source_documents"]):
            sources["sources_document"].append(document.metadata["source"])
            sources["sources_row"].append(document.metadata.get("row", None))
            sources["sources_content"].append(document.page_content)

        self.output_data["source_documents"].append(dict(sources))

    def output(self, output_path: Path):
        # output results
        date_time_now = datetime.now().strftime("%Y-%m-%d_%H-%M-%S")

        self.results_df.to_csv(
            output_path / f"answers-{date_time_now}-output.csv", index=False
        )



from mindsdb_sql.parser.ast import Show
from mindsdb.api.mongo.classes import Responder
import mindsdb.api.mongo.functions as helpers

from mindsdb.api.mongo.classes.query_sql import run_sql_command


class Responce(Responder):
    when = {'listDatabases': helpers.is_true}

    def result(self, query, request_env, mindsdb_env, session):

        ast_query = Show(category='databases')
        data = run_sql_command(request_env, ast_query)

        databases = [
            {
                'name': 'admin',
                'sizeOnDisk': 1 << 16,
                'empty': False
            }
        ]

        for row in data:
            databases.append({
                'name': list(row.values())[0],  # first value
                'sizeOnDisk': 1 << 16,
                'empty': False
            })

        return {
            'databases': databases,
            'ok': 1,
        }


responder = Responce()



"""predictor-status

Revision ID: 87b2df2b83e1
Revises: 96d5fef10caa
Create Date: 2022-09-08 14:47:45.238710

"""
import json

from alembic import op
import sqlalchemy as sa
from sqlalchemy.sql import text

import mindsdb.interfaces.storage.db    # noqa


# revision identifiers, used by Alembic.
revision = '87b2df2b83e1'
down_revision = '96d5fef10caa'
branch_labels = None
depends_on = None


def upgrade():
    with op.batch_alter_table('predictor', schema=None) as batch_op:
        batch_op.add_column(sa.Column('status', sa.String(), nullable=True))

    conn = op.get_bind()
    session = sa.orm.Session(bind=conn)
    predictors = conn.execute('''
        select id, data, update_status, json_ai, code from predictor
    ''').fetchall()

    for row in predictors:
        try:
            data = json.loads(row['data'])
        except Exception:
            data = None

        status = None
        # assume older models are complete, only temporary
        if 'status' in (data or {}):
            status = data['status']
        elif 'error' in (data or {}):
            status = 'error'
        elif row['update_status'] == 'available':
            status = 'complete'
        elif row['json_ai'] is None and row['code'] is None:
            status = 'generating'
        elif data is None:
            status = 'error'
        elif 'training_log' in (data or {}):
            status = 'training'
        elif 'error' not in (data or {}):
            status = 'complete'
        else:
            status = 'error'

        conn.execute(
            text("""
                update predictor
                set status = :status
                where id = :predictor_id
            """), {
                'status': status,
                'predictor_id': row['id']
            }
        )
    session.commit()


def downgrade():
    with op.batch_alter_table('predictor', schema=None) as batch_op:
        batch_op.drop_column('status')



"""
This module defines the wrapper for ML engines which abstracts away a lot of complexity.

In particular, three big components are included:

    - `BaseMLEngineExec` class: this class wraps any object that inherits from `BaseMLEngine` and exposes some endpoints
      normally associated with a DB handler (e.g. `native_query`, `get_tables`), as well as other ML-specific behaviors,
      like `learn()` or `predict()`. Note that while these still have to be implemented at the engine level, the burden
      on that class is lesser given that it only needs to return a pandas DataFrame. It's this class that will take said
      output and format it into the HandlerResponse instance that MindsDB core expects.

    - `learn_process` method: handles async dispatch of the `learn` method in an engine, as well as registering all
      models inside of the internal MindsDB registry.

    - `predict_process` method: handles async dispatch of the `predict` method in an engine.

"""  # noqa

import datetime as dt
import traceback
import importlib
from typing import Optional

import pandas as pd

from mindsdb_sql import parse_sql
from mindsdb_sql.parser.ast.base import ASTNode
from mindsdb_sql.parser.ast import Identifier, Select, Star, NativeQuery

from mindsdb.integrations.utilities.sql_utils import make_sql_session
from mindsdb.utilities.config import Config
import mindsdb.interfaces.storage.db as db
from mindsdb.integrations.libs.response import (
    HandlerResponse as Response,
    RESPONSE_TYPE
)
from mindsdb.__about__ import __version__ as mindsdb_version
from mindsdb.utilities.hooks import after_predict as after_predict_hook
from mindsdb.interfaces.model.model_controller import ModelController
from mindsdb.interfaces.model.functions import (
    get_model_record
)
from mindsdb.api.mysql.mysql_proxy.classes.sql_query import SQLQuery
from mindsdb.integrations.libs.const import PREDICTOR_STATUS
from mindsdb.integrations.utilities.processes import HandlerProcess
from mindsdb.utilities.functions import mark_process
from mindsdb.integrations.utilities.utils import format_exception_error
from mindsdb.interfaces.database.database import DatabaseController
from mindsdb.interfaces.storage.model_fs import ModelStorage, HandlerStorage
from mindsdb.utilities.context import context as ctx
from mindsdb.interfaces.model.functions import get_model_records

from mindsdb.integrations.handlers_client.ml_client_factory import MLClientFactory

from .ml_handler_proc import MLHandlerWrapper, MLHandlerPersistWrapper

import torch.multiprocessing as mp
mp.get_context('spawn')


class MLEngineException(Exception):
    pass


@mark_process(name='learn')
def learn_process(class_path, engine, context_dump, integration_id,
                  predictor_id, problem_definition, set_active,
                  base_predictor_id=None, training_data_df=None,
                  data_integration_ref=None, fetch_data_query=None, project_name=None):
    ctx.load(context_dump)
    db.init()

    predictor_record = db.Predictor.query.with_for_update().get(predictor_id)

    try:
        target = problem_definition['target']

        training_data_df = None

        database_controller = DatabaseController()

        sql_session = make_sql_session()
        if data_integration_ref is not None:
            if data_integration_ref['type'] == 'integration':
                integration_name = database_controller.get_integration(data_integration_ref['id'])['name']
                query = Select(
                    targets=[Star()],
                    from_table=NativeQuery(
                        integration=Identifier(integration_name),
                        query=fetch_data_query
                    )
                )
                sqlquery = SQLQuery(query, session=sql_session)
            elif data_integration_ref['type'] == 'view':
                project = database_controller.get_project(project_name)
                query_ast = parse_sql(fetch_data_query, dialect='mindsdb')
                view_query_ast = project.query_view(query_ast)
                sqlquery = SQLQuery(view_query_ast, session=sql_session)

            result = sqlquery.fetch(view='dataframe')
            training_data_df = result['result']

        training_data_columns_count, training_data_rows_count = 0, 0
        if training_data_df is not None:
            training_data_columns_count = len(training_data_df.columns)
            training_data_rows_count = len(training_data_df)

        predictor_record.training_data_columns_count = training_data_columns_count
        predictor_record.training_data_rows_count = training_data_rows_count
        db.session.commit()

        module_name, class_name = class_path
        module = importlib.import_module(module_name)
        HandlerClass = getattr(module, class_name)
        HandlerClass = MLClientFactory(handler_class=HandlerClass, engine=engine)

        handlerStorage = HandlerStorage(integration_id)
        modelStorage = ModelStorage(predictor_id)

        kwargs = {}
        if base_predictor_id is not None:
            kwargs['base_model_storage'] = ModelStorage(base_predictor_id)

        ml_handler = HandlerClass(
            engine_storage=handlerStorage,
            model_storage=modelStorage,
            **kwargs
        )

        if not ml_handler.generative:
            if training_data_df is not None and target not in training_data_df.columns:
                raise Exception(
                    f'Prediction target "{target}" not found in training dataframe: {list(training_data_df.columns)}')

        # create new model
        if base_predictor_id is None:
            ml_handler.create(target, df=training_data_df, args=problem_definition)

        # fine-tune (partially train) existing model
        else:
            # load model from previous version, use it as starting point
            problem_definition['base_model_id'] = base_predictor_id
            ml_handler.finetune(df=training_data_df, args=problem_definition)

        predictor_record.status = PREDICTOR_STATUS.COMPLETE
        predictor_record.active = set_active
        db.session.commit()
        # if retrain and set_active after success creation
        if set_active is True:
            models = get_model_records(
                name=predictor_record.name,
                project_id=predictor_record.project_id,
                active=None
            )
            for model in models:
                model.active = False
            models = [x for x in models if x.status == PREDICTOR_STATUS.COMPLETE]
            models.sort(key=lambda x: x.created_at)
            models[-1].active = True
    except Exception as e:
        print(traceback.format_exc())
        error_message = format_exception_error(e)

        predictor_record.data = {"error": error_message}
        predictor_record.status = PREDICTOR_STATUS.ERROR
        db.session.commit()

    predictor_record.training_stop_at = dt.datetime.now()
    db.session.commit()


class BaseMLEngineExec:

    def __init__(self, name, **kwargs):
        """
        ML handler interface converter
        """  # noqa
        # TODO move this class to model controller

        self.name = name
        self.config = Config()
        self.handler_controller = kwargs.get('handler_controller')
        self.company_id = kwargs.get('company_id')
        self.fs_store = kwargs.get('file_storage')
        self.storage_factory = kwargs.get('storage_factory')
        self.integration_id = kwargs.get('integration_id')
        self.execution_method = kwargs.get('execution_method')
        self.engine = kwargs.get("integration_engine")

        self.model_controller = ModelController()
        self.database_controller = DatabaseController()

        self.parser = parse_sql
        self.dialect = 'mindsdb'

        self.is_connected = True

        self.handler_class = MLClientFactory(handler_class=kwargs['handler_class'], engine=self.engine)

    def _get_ml_handler(self, predictor_id=None):
        # returns instance or wrapper over it

        integration_id = self.integration_id

        class_path = [self.handler_class.__module__, self.handler_class.__name__]

        if self.execution_method == 'subprocess':
            handler = MLHandlerWrapper()

            handler.init_handler(class_path, integration_id, predictor_id, ctx.dump())
            return handler

        elif self.execution_method == 'subprocess_keep':
            handler = MLHandlerPersistWrapper()

            handler.init_handler(class_path, integration_id, predictor_id, ctx.dump())
            return handler

        elif self.execution_method == 'remote':
            raise NotImplementedError()

        else:
            handlerStorage = HandlerStorage(integration_id)
            modelStorage = ModelStorage(predictor_id)

            ml_handler = self.handler_class(
                engine_storage=handlerStorage,
                model_storage=modelStorage,
            )
            return ml_handler

    def get_tables(self) -> Response:
        """ Returns all models currently registered that belong to the ML engine."""
        all_models = self.model_controller.get_models(integration_id=self.integration_id)
        all_models_names = [[x['name']] for x in all_models]
        response = Response(
            RESPONSE_TYPE.TABLE,
            pd.DataFrame(
                all_models_names,
                columns=['table_name']
            )
        )
        return response

    def get_columns(self, table_name: str) -> Response:
        """ Retrieves standard info about a model, e.g. data types. """  # noqa
        predictor_record = get_model_record(name=table_name, ml_handler_name=self.name)
        if predictor_record is None:
            return Response(
                RESPONSE_TYPE.ERROR,
                error_message=f"Error: model '{table_name}' does not exist!"
            )

        data = []
        if predictor_record.dtype_dict is not None:
            for key, value in predictor_record.dtype_dict.items():
                data.append((key, value))
        result = Response(
            RESPONSE_TYPE.TABLE,
            pd.DataFrame(
                data,
                columns=['COLUMN_NAME', 'DATA_TYPE']
            )
        )
        return result

    def native_query(self, query: str) -> Response:
        """ Intakes a raw SQL query and returns the answer given by the ML engine. """
        query_ast = self.parser(query, dialect=self.dialect)
        return self.query(query_ast)

    def query_(self, query: ASTNode) -> Response:
        raise Exception('Should not be used')

    def learn(
        self, model_name, project_name,
        data_integration_ref=None,
        fetch_data_query=None,
        problem_definition=None,
        join_learn_process=False,
        label=None,
        version=1,
        is_retrain=False,
        set_active=True,
    ):
        # TODO move to model_controller
        """ Trains a model given some data-gathering SQL statement. """

        target = problem_definition['target']

        project = self.database_controller.get_project(name=project_name)

        # handler-side validation
        # self.handler_class is a instance of MLClientFactory
        # so need to check self.handler_class.handler_class attribute
        # which is a class of a real MLHandler
        if hasattr(self.handler_class.handler_class, 'create_validation'):
            self.handler_class.handler_class.create_validation(target, args=problem_definition)

        predictor_record = db.Predictor(
            company_id=ctx.company_id,
            name=model_name,
            integration_id=self.integration_id,
            data_integration_ref=data_integration_ref,
            fetch_data_query=fetch_data_query,
            mindsdb_version=mindsdb_version,
            to_predict=target,
            learn_args=problem_definition,
            data={'name': model_name},
            project_id=project.id,
            training_data_columns_count=None,
            training_data_rows_count=None,
            training_start_at=dt.datetime.now(),
            status=PREDICTOR_STATUS.GENERATING,
            label=label,
            version=version,
            active=(not is_retrain),  # if create then active
        )

        db.session.add(predictor_record)
        db.session.commit()

        class_path = [self.handler_class.__module__, self.handler_class.__name__]

        p = HandlerProcess(
            learn_process,
            class_path,
            self.engine,
            ctx.dump(),
            self.integration_id,
            predictor_record.id,
            problem_definition,
            set_active,
            data_integration_ref=data_integration_ref,
            fetch_data_query=fetch_data_query,
            project_name=project_name
        )
        p.start()
        if join_learn_process is True:
            p.join()
            predictor_record = db.Predictor.query.get(predictor_record.id)
            db.session.refresh(predictor_record)

        return predictor_record

    def predict(self, model_name: str, data: list, pred_format: str = 'dict',
                project_name: str = None, version=None, params: dict = None):
        """ Generates predictions with some model and input data. """
        if isinstance(data, dict):
            data = [data]
        df = pd.DataFrame(data)
        kwargs = {
            'name': model_name,
            'ml_handler_name': self.name,
            'project_name': project_name
        }
        if version is None:
            kwargs['active'] = True
        else:
            kwargs['active'] = None
            kwargs['version'] = version
        predictor_record = get_model_record(**kwargs)
        if predictor_record is None:
            if version is not None:
                model_name = f'{model_name}.{version}'
            raise Exception(f"Error: model '{model_name}' does not exists!")
        if predictor_record.status != PREDICTOR_STATUS.COMPLETE:
            raise Exception("Error: model creation not completed")

        ml_handler = self._get_ml_handler(predictor_record.id)

        args = {
            'pred_format': pred_format,
            'predict_params': {} if params is None else params
        }
        # FIXME
        # if self.handler_class.__name__ == 'LightwoodHandler':
        if self.handler_class.__name__ == 'LightwoodHandler':
            args['code'] = predictor_record.code
            args['target'] = predictor_record.to_predict[0]
            args['dtype_dict'] = predictor_record.dtype_dict
            args['learn_args'] = predictor_record.learn_args

        if self.handler_class.__name__ in ('LangChainHandler',):
            from mindsdb.api.mysql.mysql_proxy.controllers import SessionController
            from mindsdb.api.mysql.mysql_proxy.executor.executor_commands import ExecuteCommands

            sql_session = SessionController()
            sql_session.database = 'mindsdb'

            command_executor = ExecuteCommands(sql_session, executor=None)

            args['executor'] = command_executor

        try:
            predictions = ml_handler.predict(df, args)
        except Exception as e:
            msg = str(e).strip()
            if msg == '':
                msg = e.__class__.__name__
            msg = f'[{self.name}/{model_name}]: {msg}'
            raise MLEngineException(msg) from e

        ml_handler.close()

        # mdb indexes
        if '__mindsdb_row_id' not in predictions.columns and '__mindsdb_row_id' in df.columns:
            predictions['__mindsdb_row_id'] = df['__mindsdb_row_id']

        after_predict_hook(
            company_id=self.company_id,
            predictor_id=predictor_record.id,
            rows_in_count=df.shape[0],
            columns_in_count=df.shape[1],
            rows_out_count=len(predictions)
        )
        return predictions

    def update(
            self, model_name, project_name, version,
            base_model_version: int,
            data_integration_ref=None,
            fetch_data_query=None,
            join_learn_process=False,
            label=None,
            set_active=True,
            args: Optional[dict] = None
    ):
        # generate new record from latest version as starting point
        project = self.database_controller.get_project(name=project_name)

        search_args = {
            'active': None,
            'name': model_name,
            'status': PREDICTOR_STATUS.COMPLETE
        }
        if base_model_version is not None:
            search_args['version'] = base_model_version
        predictor_records = get_model_records(**search_args)
        predictor_records.sort(key=lambda x: x.training_stop_at, reverse=True)
        predictor_records = [x for x in predictor_records if x.training_stop_at is not None]
        base_predictor_record = predictor_records[0]

        learn_args = base_predictor_record.learn_args
        learn_args['using'] = args if not learn_args.get('using', False) else {**learn_args['using'], **args}

        predictor_record = db.Predictor(
            company_id=ctx.company_id,
            name=model_name,
            integration_id=self.integration_id,
            data_integration_ref=data_integration_ref,
            fetch_data_query=fetch_data_query,
            mindsdb_version=mindsdb_version,
            to_predict=base_predictor_record.to_predict,
            learn_args=learn_args,
            data={'name': model_name},
            project_id=project.id,
            training_data_columns_count=None,
            training_data_rows_count=None,
            training_start_at=dt.datetime.now(),
            status=PREDICTOR_STATUS.GENERATING,
            label=label,
            version=version,
            active=False
        )

        db.session.add(predictor_record)
        db.session.commit()

        class_path = [self.handler_class.__module__, self.handler_class.__name__]

        p = HandlerProcess(
            learn_process,
            class_path,
            self.engine,
            ctx.dump(),
            self.integration_id,
            predictor_record.id,
            predictor_record.learn_args,
            set_active,
            base_predictor_record.id,
            data_integration_ref=data_integration_ref,
            fetch_data_query=fetch_data_query,
            project_name=project_name
        )
        p.start()
        if join_learn_process is True:
            p.join()
            predictor_record = db.Predictor.query.get(predictor_record.id)
            db.session.refresh(predictor_record)

        return predictor_record



import os
import re
import math
import json
import shutil
import tempfile
import datetime
import textwrap
import subprocess
import concurrent.futures
from typing import Optional, Dict

import openai
import numpy as np
import pandas as pd

from mindsdb.utilities import log
from mindsdb.utilities.config import Config
from mindsdb.integrations.libs.base import BaseMLEngine
from mindsdb.integrations.handlers.openai_handler.helpers import retry_with_exponential_backoff, \
    truncate_msgs_for_token_limit

CHAT_MODELS = ('gpt-3.5-turbo', 'gpt-3.5-turbo-0301', 'gpt-4', 'gpt-4-0314', 'gpt-4-32k', 'gpt-4-32k-0314')


class OpenAIHandler(BaseMLEngine):
    name = 'openai'

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.generative = True
        self.default_model = 'gpt-3.5-turbo'
        self.default_mode = 'default'  # can also be 'conversational' or 'conversational-full'
        self.supported_modes = ['default', 'conversational', 'conversational-full', 'image']
        self.rate_limit = 60  # requests per minute
        self.max_batch_size = 20
        self.default_max_tokens = 100
        self.chat_completion_models = CHAT_MODELS

    @staticmethod
    def create_validation(target, args=None, **kwargs):
        if 'using' not in args:
            raise Exception("OpenAI engine requires a USING clause! Refer to its documentation for more details.")
        else:
            args = args['using']

        if len(set(args.keys()) & {'question_column', 'prompt_template', 'json_struct', 'prompt'}) == 0:
            raise Exception('One of `question_column`, `prompt_template` or `json_struct` is required for this engine.')

        keys_collection = [
            ['prompt_template'],
            ['question_column', 'context_column'],
            ['prompt', 'user_column', 'assistant_column'],
            ['json_struct']
        ]
        for keys in keys_collection:
            if keys[0] in args and any(x[0] in args for x in keys_collection if x != keys):
                raise Exception(textwrap.dedent('''\
                    Please provide one of
                        1) a `prompt_template`
                        2) a `question_column` and an optional `context_column`
                        3) a `json_struct`
                        4) a `prompt' and 'user_column' and 'assistant_column`
                '''))

    def create(self, target, args=None, **kwargs):
        args = args['using']

        args['target'] = target
        available_models = [m.openai_id for m in openai.Model.list(api_key=self._get_openai_api_key(args)).data]
        if not args.get('model_name'):
            args['model_name'] = self.default_model
        elif args['model_name'] not in available_models:
            raise Exception(f"Invalid model name. Please use one of {available_models}")

        if not args.get('mode'):
            args['mode'] = self.default_mode
        elif args['mode'] not in self.supported_modes:
            raise Exception(f"Invalid operation mode. Please use one of {self.supported_modes}")

        self.model_storage.json_set('args', args)

    def _get_openai_api_key(self, args, strict=True):
        """ 
        API_KEY preference order:
            1. provided at model creation
            2. provided at engine creation
            3. OPENAI_API_KEY env variable
            4. openai.api_key setting in config.json
        """  # noqa
        # 1
        if 'api_key' in args:
            return args['api_key']
        # 2
        connection_args = self.engine_storage.get_connection_args()
        if 'api_key' in connection_args:
            return connection_args['api_key']
        # 3
        api_key = os.getenv('OPENAI_API_KEY')
        if api_key is not None:
            return api_key
        # 4
        config = Config()
        openai_cfg = config.get('openai', {})
        if 'api_key' in openai_cfg:
            return openai_cfg['api_key']

        if strict:
            raise Exception(f'Missing API key "api_key". Either re-create this ML_ENGINE specifying the `api_key` parameter,\
                 or re-create this model and pass the API key with `USING` syntax.')  # noqa

    def predict(self, df, args=None):
        """
        If there is a prompt template, we use it. Otherwise, we use the concatenation of `context_column` (optional) and `question_column` to ask for a completion.
        """ # noqa
        # TODO: support for edits, embeddings and moderation

        pred_args = args['predict_params'] if args else {}
        args = self.model_storage.json_get('args')
        df = df.reset_index(drop=True)

        if pred_args.get('mode'):
            if pred_args['mode'] in self.supported_modes:
                args['mode'] = pred_args['mode']
            else:
                raise Exception(f"Invalid operation mode. Please use one of {self.supported_modes}.")  # noqa

        if pred_args.get('prompt_template', False):
            base_template = pred_args['prompt_template']  # override with predict-time template if available
        elif args.get('prompt_template', False):
            base_template = args['prompt_template']
        else:
            base_template = None

        # Image mode
        if args.get('mode', self.default_mode) == 'image':
            api_args = {
                'n': pred_args.get('n', None),
                'size': pred_args.get('size', None),
                'response_format': pred_args.get('response_format', None),
            }
            api_args = {k: v for k, v in api_args.items() if v is not None}  # filter out non-specified api args
            model_name = 'image'

            if args.get('question_column'):
                prompts = list(df[args['question_column']].apply(lambda x: str(x)))
                empty_prompt_ids = np.where(df[[args['question_column']]].isna().all(axis=1).values)[0]
            elif args.get('prompt_template'):
                prompts, empty_prompt_ids = self._get_completed_prompts(base_template, df)
            else:
                raise Exception('Image mode needs either `prompt_template` or `question_column`.')

        # Chat or normal completion mode
        else:
            if args.get('question_column', False) and args['question_column'] not in df.columns:
                raise Exception(f"This model expects a question to answer in the '{args['question_column']}' column.")

            if args.get('context_column', False) and args['context_column'] not in df.columns:
                raise Exception(f"This model expects context in the '{args['context_column']}' column.")

            # api argument validation
            model_name = args.get('model_name', self.default_model)
            api_args = {
                'max_tokens': pred_args.get('max_tokens', args.get('max_tokens', self.default_max_tokens)),
                'temperature': min(1.0, max(0.0, pred_args.get('temperature', args.get('temperature', 0.0)))),
                'top_p': pred_args.get('top_p', None),
                'n': pred_args.get('n', None),
                'stop': pred_args.get('stop', None),
                'presence_penalty': pred_args.get('presence_penalty', None),
                'frequency_penalty': pred_args.get('frequency_penalty', None),
                'best_of': pred_args.get('best_of', None),
                'logit_bias': pred_args.get('logit_bias', None),
                'user': pred_args.get('user', None),
            }

            if args.get('mode', self.default_mode) != 'default' and model_name not in self.chat_completion_models:
                raise Exception(f"Conversational modes are only available for the following models: {', '.join(self.chat_completion_models)}")  # noqa

            if args.get('prompt_template', False):
                prompts, empty_prompt_ids = self._get_completed_prompts(base_template, df)

            elif args.get('context_column', False):
                empty_prompt_ids = np.where(df[[args['context_column'],
                                               args['question_column']]].isna().all(axis=1).values)[0]
                contexts = list(df[args['context_column']].apply(lambda x: str(x)))
                questions = list(df[args['question_column']].apply(lambda x: str(x)))
                prompts = [f'Context: {c}\nQuestion: {q}\nAnswer: ' for c, q in zip(contexts, questions)]

            elif args.get('json_struct', False):
                empty_prompt_ids = np.where(df[[args['input_text']]].isna().all(axis=1).values)[0]
                prompts = []
                for i in df.index:
                    if 'json_struct' in df.columns:
                        if isinstance(df['json_struct'][i], str):
                            df['json_struct'][i] = json.loads(df['json_struct'][i])
                        json_struct = ''
                        for ind, val in enumerate(df['json_struct'][i].values()):
                            json_struct = json_struct + f'{ind}. {val}\n'
                    else:
                        json_struct = ''
                        for ind, val in enumerate(args['json_struct'].values()):
                            json_struct = json_struct + f'{ind + 1}. {val}\n'

                    p = textwrap.dedent(f'''\
                        Using text starting after 'The text is:', give exactly {len(args['json_struct'])} answers to the questions:
                        {{{{json_struct}}}}
    
                        Answers should be in the same order as the questions.
                        Each answer should start with a question number.
                        Each answer must end with new line.
                        If there is no answer to the question in the text, put a -.
                        Answers should be as short as possible, ideally 1-2 words (unless otherwise specified).
    
                        The text is:
                        {{{{{args['input_text']}}}}}
                    ''')
                    p = p.replace('{{json_struct}}', json_struct)
                    for column in df.columns:
                        if column == 'json_struct':
                            continue
                        p = p.replace(f'{{{{{column}}}}}', str(df[column][i]))
                    prompts.append(p)
            elif 'prompt' in args:
                empty_prompt_ids = []
                prompts = list(df[args['user_column']])
            else:
                empty_prompt_ids = np.where(df[[args['question_column']]].isna().all(axis=1).values)[0]
                prompts = list(df[args['question_column']].apply(lambda x: str(x)))

        # remove prompts without signal from completion queue
        prompts = [j for i, j in enumerate(prompts) if i not in empty_prompt_ids]

        api_key = self._get_openai_api_key(args)
        api_args = {k: v for k, v in api_args.items() if v is not None}  # filter out non-specified api args
        completion = self._completion(model_name, prompts, api_key, api_args, args, df)

        # add null completion for empty prompts
        for i in sorted(empty_prompt_ids):
            completion.insert(i, None)

        pred_df = pd.DataFrame(completion, columns=[args['target']])

        # restore json struct
        if args.get('json_struct', False):
            for i in pred_df.index:
                try:
                    if 'json_struct' in df.columns:
                        json_keys = df['json_struct'][i].keys()
                    else:
                        json_keys = args['json_struct'].keys()
                    responses = pred_df[args['target']][i].split('\n')
                    responses = [x[3:] for x in responses]      # del question index

                    pred_df[args['target']][i] = {
                        key: val for key, val in zip(
                            json_keys,
                            responses
                        )
                    }
                except Exception:
                    pred_df[args['target']][i] = None

        return pred_df

    def _completion(self, model_name, prompts, api_key, api_args, args, df, parallel=True):
        """
        Handles completion for an arbitrary amount of rows.

        There are a couple checks that should be done when calling OpenAI's API:
          - account max batch size, to maximize batch size first
          - account rate limit, to maximize parallel calls second

        Additionally, single completion calls are done with exponential backoff to guarantee all prompts are processed,
        because even with previous checks the tokens-per-minute limit may apply.
        """
        @retry_with_exponential_backoff()
        def _submit_completion(model_name, prompts, api_key, api_args, args, df):
            kwargs = {
                'model': model_name,
                'api_key': api_key,
                'organization': args.get('api_organization'),
            }
            if model_name == 'image':
                return _submit_image_completion(kwargs, prompts, api_args)
            elif model_name in self.chat_completion_models:
                return _submit_chat_completion(kwargs, prompts, api_args, df, mode=args.get('mode', 'conversational'))
            else:
                return _submit_normal_completion(kwargs, prompts, api_args)

        def _log_api_call(params, response):
            params2 = params.copy()
            params2.pop('api_key', None)
            params2.pop('user', None)
            log.logger.debug(f'>>>openai call: {params2}:\n{response}')

        def _submit_normal_completion(kwargs, prompts, api_args):
            def _tidy(comp):
                tidy_comps = []
                for c in comp['choices']:
                    if 'text' in c:
                        tidy_comps.append(c['text'].strip('\n').strip(''))
                return tidy_comps

            kwargs['prompt'] = prompts
            kwargs = {**kwargs, **api_args}

            resp = _tidy(openai.Completion.create(**kwargs))
            _log_api_call(kwargs, resp)
            return resp

        def _submit_chat_completion(kwargs, prompts, api_args, df, mode='conversational'):
            def _tidy(comp):
                tidy_comps = []
                for c in comp['choices']:
                    if 'message' in c:
                        tidy_comps.append(c['message']['content'].strip('\n').strip(''))
                return tidy_comps

            completions = []
            if mode != 'conversational':
                initial_prompt = {"role": "system", "content": "You are a helpful assistant. Your task is to continue the chat."}  # noqa
            else:
                # get prompt from model
                initial_prompt = {"role": "system",  "content": args['prompt']}  # noqa

            kwargs['messages'] = [initial_prompt]
            last_completion_content = None

            for pidx in range(len(prompts)):
                if mode != 'conversational':
                    kwargs['messages'].append({'role': 'user', 'content': prompts[pidx]})
                else:
                    question = prompts[pidx]
                    if question:
                        kwargs['messages'].append({'role': 'user', 'content': question})
                    answer = df.iloc[pidx][args.get('assistant_column')]
                    if answer:
                        kwargs['messages'].append({'role': 'assistant', 'content': answer})

                if mode == 'conversational-full' or (mode == 'conversational' and pidx == len(prompts) - 1):
                    kwargs['messages'] = truncate_msgs_for_token_limit(kwargs['messages'],
                                                                       kwargs['model'],
                                                                       api_args['max_tokens'])
                    pkwargs = {**kwargs, **api_args}
                    resp = _tidy(openai.ChatCompletion.create(**pkwargs))
                    _log_api_call(pkwargs, resp)

                    completions.extend(resp)
                elif mode == 'default':
                    kwargs['messages'] = [initial_prompt] + [kwargs['messages'][-1]]
                    pkwargs = {**kwargs, **api_args}

                    resp = _tidy(openai.ChatCompletion.create(**pkwargs))
                    _log_api_call(pkwargs, resp)

                    completions.extend(resp)
                else:
                    # in "normal" conversational mode, we request completions only for the last row
                    last_completion_content = None
                    if args.get('answer_column') in df.columns:
                        # insert completion if provided, which saves redundant API calls
                        completions.extend([df.iloc[pidx][args.get('answer_column')]])
                    else:
                        completions.extend([''])

                if args.get('answer_column') in df.columns:
                    kwargs['messages'].append({'role': 'assistant',
                                               'content': df.iloc[pidx][args.get('answer_column')]})
                elif last_completion_content:
                    # interleave assistant responses with user input
                    kwargs['messages'].append({'role': 'assistant', 'content': last_completion_content[0]})

            return completions

        def _submit_image_completion(kwargs, prompts, api_args):
            def _tidy(comp):
                return [c[0]['url'] if 'url' in c[0].keys() else c[0]['b64_json'] for c in comp]
            kwargs.pop('model')
            completions = [openai.Image.create(**{'prompt': p, **kwargs, **api_args})['data'] for p in prompts]
            return _tidy(completions)

        try:
            # check if simple completion works
            completion = _submit_completion(
                model_name,
                prompts,
                api_key,
                api_args,
                args,
                df
            )
            return completion
        except openai.error.InvalidRequestError as e:
            # else, we get the max batch size
            e = e.user_message
            if 'you can currently request up to at most a total of' in e:
                pattern = 'a total of'
                max_batch_size = int(e[e.find(pattern) + len(pattern):].split(').')[0])
            else:
                max_batch_size = self.max_batch_size  # guards against changes in the API message

        if not parallel:
            completion = None
            for i in range(math.ceil(len(prompts) / max_batch_size)):
                partial = _submit_completion(model_name,
                                             prompts[i * max_batch_size:(i + 1) * max_batch_size],
                                             api_key,
                                             api_args,
                                             args,
                                             df)
                if not completion:
                    completion = partial
                else:
                    completion['choices'].extend(partial['choices'])
                    for field in ('prompt_tokens', 'completion_tokens', 'total_tokens'):
                        completion['usage'][field] += partial['usage'][field]
        else:
            promises = []
            with concurrent.futures.ThreadPoolExecutor() as executor:
                for i in range(math.ceil(len(prompts) / max_batch_size)):
                    print(f'{i * max_batch_size}:{(i+1) * max_batch_size}/{len(prompts)}')
                    future = executor.submit(_submit_completion,
                                             model_name,
                                             prompts[i * max_batch_size:(i + 1) * max_batch_size],
                                             api_key,
                                             api_args,
                                             args,
                                             df)
                    promises.append({"choices": future})
            completion = None
            for p in promises:
                if not completion:
                    completion = p['choices'].result()
                else:
                    completion['choices'].extend(p['choices'].result()['choices'])

        return completion

    def describe(self, attribute: Optional[str] = None) -> pd.DataFrame:
        # TODO: Update to use update() artifacts

        args = self.model_storage.json_get('args')

        if attribute == 'args':
            return pd.DataFrame(args.items(), columns=['key', 'value'])
        elif attribute == 'metadata':
            api_key = self._get_openai_api_key(args)
            model_name = args.get('model_name', self.default_model)
            meta = openai.Model.retrieve(model_name, api_key=api_key)
            return pd.DataFrame(meta.items(), columns=['key', 'value'])
        else:
            tables = ['args', 'metadata']
            return pd.DataFrame(tables, columns=['tables'])

    def finetune(self, df: Optional[pd.DataFrame] = None, args: Optional[Dict] = None) -> None:
        """
        Fine-tune OpenAI GPT models. Steps are roughly:
          - Analyze input data and modify it according to suggestions made by the OpenAI utility tool
          - Get a training and validation file
          - Determine base model to use
          - Submit a fine-tuning job via the OpenAI API
          - Monitor progress with exponential backoff (which has been modified for greater control given a time budget in hours), 
          - Gather stats once fine-tuning finishes
          - Modify model metadata so that the new version triggers the fine-tuned version of the model (stored in the user's OpenAI account)

        Caveats: 
          - As base fine-tuning models, OpenAI only supports the original GPT ones: `ada`, `babbage`, `curie`, `davinci`. This means if you fine-tune successively more than once, any fine-tuning other than the most recent one is lost.
        """  # noqa

        args = args if args else {}
        using_args = args.pop('using') if 'using' in args else {}
        prompt_col = using_args.get('prompt_column', 'prompt')
        completion_col = using_args.get('completion_column', 'completion')

        for col in [prompt_col, completion_col]:
            if col not in set(df.columns):
                raise Exception(f"To fine-tune this OpenAI model, please format your select data query to have a `{prompt_col}` column and a `{completion_col}` column first.")  # noqa

        args = {**using_args, **args}
        prev_model_name = self.base_model_storage.json_get('args').get('model_name', '')

        openai.api_key = self._get_openai_api_key(args)
        finetune_time = datetime.datetime.now().strftime('%Y-%m-%d_%H-%M-%S')

        temp_storage_path = tempfile.mkdtemp()
        temp_file_name = f"ft_{finetune_time}"
        temp_model_storage_path = f"{temp_storage_path}/{temp_file_name}.jsonl"
        df.to_json(temp_model_storage_path, orient='records', lines=True)

        # TODO avoid subprocess usage once OpenAI enables non-CLI access
        subprocess.run(
            [
                "openai", "tools", "fine_tunes.prepare_data",
                "-f", temp_model_storage_path,                  # from file
                '-q'                                            # quiet mode (accepts all suggestions)
            ],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            encoding="utf-8",
        )

        file_names = {'original': f'{temp_file_name}.jsonl',
                      'base': f'{temp_file_name}_prepared.jsonl',
                      'train': f'{temp_file_name}_prepared_train.jsonl',
                      'val': f'{temp_file_name}_prepared_valid.jsonl'}
        jsons = {k: None for k in file_names.keys()}
        for split, file_name in file_names.items():
            if os.path.isfile(os.path.join(temp_storage_path, file_name)):
                jsons[split] = openai.File.create(
                    file=open(f"{temp_storage_path}/{file_name}", "rb"),
                    purpose='fine-tune')

        train_file_id = jsons['train'].id if isinstance(jsons['train'], openai.File) else jsons['base']
        val_file_id = jsons['val'].id if isinstance(jsons['val'], openai.File) else None

        def _get_model_type(model_name: str):
            for model_type in ['ada', 'curie', 'babbage', 'davinci']:
                if model_type in model_name.lower():
                    return model_type
            return 'ada'

        # `None` values are internally imputed by OpenAI to `null` or default values
        ft_params = {
            'training_file': train_file_id,
            'validation_file': val_file_id,
            'model': _get_model_type(prev_model_name),
            'suffix': 'mindsdb',
            'n_epochs': using_args.get('n_epochs', None),
            'batch_size': using_args.get('batch_size', None),
            'learning_rate_multiplier': using_args.get('learning_rate_multiplier', None),
            'prompt_loss_weight': using_args.get('prompt_loss_weight', None),
            'compute_classification_metrics': using_args.get('compute_classification_metrics', None),
            'classification_n_classes': using_args.get('classification_n_classes', None),
            'classification_positive_class': using_args.get('classification_positive_class', None),
            'classification_betas': using_args.get('classification_betas', None),
        }

        start_time = datetime.datetime.now()
        ft_result = openai.FineTune.create(**{k: v for k, v in ft_params.items() if v is not None})

        @retry_with_exponential_backoff(
            hour_budget=args.get('hour_budget', 8),
            errors=(openai.error.RateLimitError, openai.error.OpenAIError))
        def _check_ft_status(model_id):
            ft_retrieved = openai.FineTune.retrieve(id=model_id)
            if ft_retrieved['status'] in ('succeeded', 'failed'):
                return ft_retrieved
            else:
                raise openai.error.OpenAIError('Fine-tuning still pending!')

        ft_stats = _check_ft_status(ft_result.id)
        ft_model_name = ft_stats['fine_tuned_model']

        if ft_stats['status'] != 'succeeded':
            raise Exception(f"Fine-tuning did not complete successfully (status: {ft_stats['status']}). Error message: {ft_stats['events'][-1]['message']}")  # noqa

        end_time = datetime.datetime.now()
        runtime = end_time - start_time

        result_file_id = openai.FineTune.retrieve(id=ft_result.id)['result_files'][0].id
        name_extension = openai.File.retrieve(id=result_file_id).filename
        result_path = f'{temp_storage_path}/ft_{finetune_time}_result_{name_extension}'
        with open(result_path, 'wb') as f:
            f.write(openai.File.download(id=result_file_id))

        train_stats = pd.read_csv(result_path)
        if 'validation_token_accuracy' in train_stats.columns:
            train_stats = train_stats[train_stats['validation_token_accuracy'].notnull()]

        args['model_name'] = ft_model_name
        args['ft_api_info'] = ft_stats.to_dict_recursive()
        args['ft_result_stats'] = train_stats.to_dict()
        args['runtime'] = runtime.total_seconds()
        args['mode'] = self.base_model_storage.json_get('args').get('mode', self.default_mode)

        self.model_storage.json_set('args', args)
        shutil.rmtree(temp_storage_path)

    @staticmethod
    def _get_completed_prompts(base_template, df):
        columns = []
        spans = []
        matches = list(re.finditer("{{(.*?)}}", base_template))

        assert len(matches) > 0, 'No placeholders found in the prompt, please provide a valid prompt template.'

        first_span = matches[0].start()
        last_span = matches[-1].end()

        for m in matches:
            columns.append(m[0].replace('{', '').replace('}', ''))
            spans.extend((m.start(), m.end()))

        spans = spans[1:-1]  # omit first and last, they are added separately
        template = [base_template[s:e] for s, e in list(zip(spans, spans[1:]))[::2]]  # take every other to skip placeholders  # noqa
        template.insert(0, base_template[0:first_span])  # add prompt start
        template.append(base_template[last_span:])  # add prompt end

        empty_prompt_ids = np.where(df[columns].isna().all(axis=1).values)[0]

        df['__mdb_prompt'] = ''
        for i in range(len(template)):
            atom = template[i]
            if i < len(columns):
                col = df[columns[i]].replace(to_replace=[None], value='')  # add empty quote if data is missing
                df['__mdb_prompt'] = df['__mdb_prompt'].apply(lambda x: x + atom) + col.astype("string")
            else:
                df['__mdb_prompt'] = df['__mdb_prompt'].apply(lambda x: x + atom)
        prompts = list(df['__mdb_prompt'])

        return prompts, empty_prompt_ids



import json
import requests
from typing import Dict, List


class FrappeClient(object):
    """Client to interact with the Frappe API.
    
    Attributes:
        domain (str): Path to Frappe domain to use (e.g. https://mindsdbfrappe.com).
        access_token (str): Frappe authorization token to use for all API requests.
    """

    def __init__(
            self,
            domain: str,
            access_token: str):
        self.domain = domain
        self.base_url = f'{self.domain}/api'
        self.access_token = access_token

        self.headers = {
            'Authorization': f'token {self.access_token}',
        }

    def get_document(self, doctype: str, name: str) -> Dict:
        """Gets a document matching the given doctype from Frappe.
        
        See https://frappeframework.com/docs/v14/user/en/api/rest#listing-documents
        Args:
            doctype (str): The document type to retrieve.
            name (str): Name of the document.
        """
        document_response = requests.get(
            f'{self.base_url}/resource/{doctype}/{name}',
            headers=self.headers)
        if not document_response.ok:
            document_response.raise_for_status()
        return document_response.json()['data']

    def get_documents(self, doctype: str, limit: int = None, fields: List[str] = None, filters: List[List] = None) -> List[Dict]:
        """Gets all documents matching the given doctype from Frappe.
        
        See https://frappeframework.com/docs/v14/user/en/api/rest#listing-documents
        Args:
            doctype (str): The document type to retrieve.
            limit (int): At most, how many messages to return.
            filters (List[List]): List of filters in the form [field, operator, value] e.g. ["amount", ">", 50]
        """
        params = {
            'fields': json.dumps(["*"])
        }
        if limit is not None:
            params['limit_page_length'] = limit
        if filters is not None:
            params['filters'] = json.dumps(filters)
        if fields is not None:
            params['fields'] = json.dumps(fields)
        documents_response = requests.get(
            f'{self.base_url}/resource/{doctype}/',
            params=params,
            headers=self.headers,
            allow_redirects=False)
        if documents_response.is_redirect:
            # We have to manually redirect to preserve the 'Authorization' header.
            # See https://github.com/request/request/pull/1184/commits/210b326fd8625f358e06c59dc11e74468b1de515.
            redirect_url = documents_response.headers.get('location', None)
            if redirect_url is None:
                raise requests.HTTPError('Could not find redirect URL')
            documents_response = requests.get(
                redirect_url,
                params=params,
                headers=self.headers,
                allow_redirects=False)

        if not documents_response.ok:
            documents_response.raise_for_status()
        return documents_response.json()['data']

    def post_document(
            self,
            doctype: str,
            data: Dict):
        """Creates a new document of the given doctype.
        See https://frappeframework.com/docs/v14/user/en/api/rest#listing-documents
        
        Args:
            doctype (str): Type of the document to create.
            data (Dict): Document object.
        """
        post_response = requests.post(
            f'{self.base_url}/resource/{doctype}',
            json=data,
            headers=self.headers)
        if not post_response.ok:
            post_response.raise_for_status()
        return post_response.json()['data']

    def ping(self) -> bool:
        """Sends a basic request to the Frappe API to see if it succeeds.
        
        Returns whether or not the connection to the Frappe API is valid.
        See https://frappeframework.com/docs/v14/user/en/api/rest#1-token-based-authentication
        """

        # No ping or similar endpoint exists, so we'll try getting the logged in user.
        user_response = requests.get(
            f'{self.base_url}/method/frappe.auth.get_logged_user',
            headers=self.headers)
        return user_response.ok



import os
from mindsdb.integrations.libs.handler_helpers import discover_services
from mindsdb.utilities.log import get_log


logger = get_log(logger_name="main")


class MLClientFactory:
    def __init__(self, handler_class, engine):
        self.engine = engine
        self.client_class = None
        # self.client_class = MLClientGRPC
        self.handler_class = handler_class
        self.__name__ = self.handler_class.__name__
        self.__module__ = self.handler_class.__module__

    def __call__(self, engine_storage, model_storage, **kwargs):
        service_info = self.discover_service(self.engine)
        if service_info:
            host = service_info["host"]
            port = service_info["port"]
        else:
            host = os.environ.get("MINDSDB_ML_SERVICE_HOST", None)
            port = os.environ.get("MINDSDB_ML_SERVICE_PORT", None)

        if host is not None and port is not None:
            try:
                from mindsdb.integrations.handlers_client.ml_grpc_client import MLClientGRPC
                self.client_class = MLClientGRPC
            except (ImportError, ModuleNotFoundError):
                logger.error("to use microservice mode please install 'pip install mindsdb[grpc]'")
                host = None
                port = None

        if host is None or port is None:
            logger.info(
                "%s.__call__: no post/port to MLService have provided. Handle all ML request locally",
                self.__class__.__name__,
            )
            return self.handler_class(engine_storage=engine_storage, model_storage=model_storage, **kwargs)

        logger.info("%s.__call__: api to communicate with ML services - gRPC, host - %s, port - %s",
                    self.__class__.__name__,
                    host,
                    port,
                    )
        kwargs["engine"] = self.engine
        return self.client_class(host, port, integration_id=engine_storage.integration_id, predictor_id=model_storage.predictor_id, **kwargs)

    def discover_service(self, engine):
        discover_url = os.environ.get("REGISTRY_URL")
        if not discover_url:
            return {}
        discover_url = f"{discover_url}/discover"
        res = discover_services(discover_url)
        if not res:
            return {}
        return res[engine][0]
